#
# Unit Tests for Suffix
#

import os
import sys
import copy
import itertools
import pickle
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

from coopr.pyomo import *
from coopr.pyomo.base.suffix import active_export_suffix_generator, \
                                    export_suffix_generator, \
                                    active_import_suffix_generator, \
                                    import_suffix_generator, \
                                    active_local_suffix_generator, \
                                    local_suffix_generator, \
                                    active_suffix_generator, \
                                    suffix_generator
import pyutilib.th as unittest

from six import StringIO

def simple_con_rule(model,i):
    return model.x[i] == 1
def simple_obj_rule(model,i):
    return model.x[i]

class TestSuffixMethods(unittest.TestCase):
    
    # test __init__
    def test_init(self):
        model = ConcreteModel()
        # no keywords
        model.junk = Suffix()
        model.del_component('junk')
        
        for direction,datatype in itertools.product(Suffix.SuffixDirections,Suffix.SuffixDatatypes):
            model.junk = Suffix(direction=direction,datatype=datatype)
            model.del_component('junk')
            
    # test importEnabled
    def test_importEnabled(self):
        model = ConcreteModel()
        model.test_local = Suffix(direction=Suffix.LOCAL)
        self.assertTrue(model.test_local.importEnabled() is False)

        model.test_out = Suffix(direction=Suffix.IMPORT)
        self.assertTrue(model.test_out.importEnabled() is True)

        model.test_in = Suffix(direction=Suffix.EXPORT)
        self.assertTrue(model.test_in.importEnabled() is False)

        model.test_inout = Suffix(direction=Suffix.IMPORT_EXPORT)        
        self.assertTrue(model.test_inout.importEnabled() is True)

    # test exportEnabled
    def test_exportEnabled(self):
        model = ConcreteModel()

        model.test_local = Suffix(direction=Suffix.LOCAL)        
        self.assertTrue(model.test_local.exportEnabled() is False)

        model.test_out = Suffix(direction=Suffix.IMPORT)
        self.assertTrue(model.test_out.exportEnabled() is False)

        model.test_in = Suffix(direction=Suffix.EXPORT)
        self.assertTrue(model.test_in.exportEnabled() is True)

        model.test_inout = Suffix(direction=Suffix.IMPORT_EXPORT)
        self.assertTrue(model.test_inout.exportEnabled() is True)

    # test setValue with wrong number of args
    def test_setValue_wrongnumberargs(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        try:
            model.junk.setValue()
        except TypeError:
            pass
        else:
            self.fail("Trying to call setValue with wrong number of arguments should fail.")
        try:
            model.junk.setValue(1,2,3)
        except TypeError:
            pass
        else:
            self.fail("Trying to call setValue with wrong number of arguments should fail.")

    # test setValue and getValue
    # and if Var arrays are correctly expanded
    def test_setValue_getValue_Var(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1,2,3])

        model.junk.setValue(model.X,1.0)
        model.junk.setValue(model.X[1],2.0)

        self.assertEqual(model.junk.getValue(model.X), None)
        self.assertEqual(model.junk.getValue(model.X[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.X[2]), 1.0)

        self.assertEqual(model.junk.getValue(model.x), None)

        model.junk.setValue(model.x,3.0)
        model.junk.setValue(model.X[2],3.0)

        self.assertEqual(model.junk.getValue(model.X), None)
        self.assertEqual(model.junk.getValue(model.X[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.X[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.x), 3.0)

    # test setValue and getValue
    # and if Constraint arrays are correctly expanded
    def test_setValue_getValue_Constraint(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1,2,3])
        model.c = Constraint(expr=model.x>=1)
        model.C = Constraint([1,2,3], rule=lambda model,i: model.X[i]>=1)

        model.junk.setValue(model.C,1.0)
        model.junk.setValue(model.C[1],2.0)

        self.assertEqual(model.junk.getValue(model.C), None)
        self.assertEqual(model.junk.getValue(model.C[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.C[2]), 1.0)

        self.assertEqual(model.junk.getValue(model.c), None)

        model.junk.setValue(model.c,3.0)
        model.junk.setValue(model.C[2],3.0)

        self.assertEqual(model.junk.getValue(model.C), None)
        self.assertEqual(model.junk.getValue(model.C[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.C[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.c), 3.0)

    # test setValue and getValue
    # and if Objective arrays are correctly expanded
    def test_setValue_getValue_Objective(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1,2,3])
        model.obj = Objective(expr=summation(model.X)+model.x)
        model.OBJ = Objective([1,2,3], rule=lambda model,i: model.X[i])

        model.junk.setValue(model.OBJ,1.0)
        model.junk.setValue(model.OBJ[1],2.0)

        self.assertEqual(model.junk.getValue(model.OBJ), None)
        self.assertEqual(model.junk.getValue(model.OBJ[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.OBJ[2]), 1.0)

        self.assertEqual(model.junk.getValue(model.obj), None)

        model.junk.setValue(model.obj,3.0)
        model.junk.setValue(model.OBJ[2],3.0)

        self.assertEqual(model.junk.getValue(model.OBJ), None)
        self.assertEqual(model.junk.getValue(model.OBJ[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.OBJ[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.obj), 3.0)

    # test setValue and getValue
    # and if mutable Param arrays are correctly expanded
    def Xtest_setValue_getValue_mutableParam(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1,2,3])
        model.p = Param(initialize=1.0,mutable=True)
        model.P = Param([1,2,3],initialize=1.0,mutable=True)

        model.junk.setValue(model.P,1.0)
        model.junk.setValue(model.P[1],2.0)

        self.assertEqual(model.junk.getValue(model.P), None)
        self.assertEqual(model.junk.getValue(model.P[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.P[2]), 1.0)

        self.assertEqual(model.junk.getValue(model.p), None)

        model.junk.setValue(model.p,3.0)
        model.junk.setValue(model.P[2],3.0)

        self.assertEqual(model.junk.getValue(model.P), None)
        self.assertEqual(model.junk.getValue(model.P[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.P[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.p), 3.0)

    # test setValue and getValue
    # and if immutable Param arrays are correctly expanded
    def Xtest_setValue_getValue_immutableParam(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1,2,3])
        model.p = Param(initialize=1.0,mutable=False)
        model.P = Param([1,2,3],initialize=1.0,mutable=False)

        model.junk.setValue(model.P,1.0)
        model.junk.setValue(model.P[1],2.0)

        self.assertEqual(model.junk.getValue(model.P), None)
        self.assertEqual(model.junk.getValue(model.P[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.P[2]), 1.0)

        self.assertEqual(model.junk.getValue(model.p), None)

        model.junk.setValue(model.p,3.0)
        model.junk.setValue(model.P[2],3.0)

        self.assertEqual(model.junk.getValue(model.P), None)
        self.assertEqual(model.junk.getValue(model.P[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.P[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.p), 3.0)

    # test setValue and getValue
    # and if Set arrays are correctly expanded
    def Xtest_setValue_getValue_Set(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.X = Var([1,2,3])
        model.s = Set(initialize=[1,2,3])
        model.S = Set([1,2,3],initialize={1:[1,2,3],2:[1,2,3],3:[1,2,3]})

        model.junk.setValue(model.S,1.0)
        model.junk.setValue(model.S[1],2.0)

        self.assertEqual(model.junk.getValue(model.S), None)
        self.assertEqual(model.junk.getValue(model.S[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.S[2]), 1.0)

        self.assertEqual(model.junk.getValue(model.s), None)

        model.junk.setValue(model.s,3.0)
        model.junk.setValue(model.S[2],3.0)

        self.assertEqual(model.junk.getValue(model.S), None)
        self.assertEqual(model.junk.getValue(model.S[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.S[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.s), 3.0)

    # test setValue and getValue
    # and if Block arrays are correctly expanded
    def test_setValue_getValue_Block(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.b = Block()
        model.B = Block([1,2,3])

        # make sure each BlockData gets construced
        model.B[1].x = 1
        model.B[2].x = 2
        model.B[3].x = 3

        model.junk.setValue(model.B,1.0)
        model.junk.setValue(model.B[1],2.0)

        self.assertEqual(model.junk.getValue(model.B), None)
        self.assertEqual(model.junk.getValue(model.B[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.B[2]), 1.0)

        self.assertEqual(model.junk.getValue(model.b), None)

        model.junk.setValue(model.b,3.0)
        model.junk.setValue(model.B[2],3.0)

        self.assertEqual(model.junk.getValue(model.B), None)
        self.assertEqual(model.junk.getValue(model.B[1]), 2.0)
        self.assertEqual(model.junk.getValue(model.B[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.b), 3.0)

    # test setValue with no component argument
    def test_setAllValues(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1,2,3])
        model.z = Var([1,2,3])

        model.junk.setValue(model.y[2],1.0)
        model.junk.setValue(model.z,2.0)
        
        self.assertTrue(model.junk.getValue(model.x) is None)
        self.assertTrue(model.junk.getValue(model.y) is None)
        self.assertTrue(model.junk.getValue(model.y[1]) is None)
        self.assertEqual(model.junk.getValue(model.y[2]), 1.0)
        self.assertEqual(model.junk.getValue(model.z), None)
        self.assertEqual(model.junk.getValue(model.z[1]), 2.0)
        
        model.junk.setAllValues(3.0)

        self.assertTrue(model.junk.getValue(model.x) is None)
        self.assertTrue(model.junk.getValue(model.y) is None)
        self.assertTrue(model.junk.getValue(model.y[1]) is None)
        self.assertEqual(model.junk.getValue(model.y[2]), 3.0)
        self.assertEqual(model.junk.getValue(model.z), None)
        self.assertEqual(model.junk.getValue(model.z[1]), 3.0)

    # test updateValues
    def test_updateValues(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.junk.setValue(model.x,0.0)
        self.assertEqual(model.junk.getValue(model.x),0.0)
        self.assertEqual(model.junk.getValue(model.y),None)
        self.assertEqual(model.junk.getValue(model.z),None)
        model.junk.updateValues([(model.x,1.0),(model.y,2.0),(model.z,3.0)])
        self.assertEqual(model.junk.getValue(model.x),1.0)
        self.assertEqual(model.junk.getValue(model.y),2.0)
        self.assertEqual(model.junk.getValue(model.z),3.0)


    # test clearValue 
    def test_clearValue(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1,2,3])
        model.z = Var([1,2,3])

        model.junk.setValue(model.x,-1.0)
        model.junk.setValue(model.y,-2.0)
        model.junk.setValue(model.y[2],1.0)
        model.junk.setValue(model.z,2.0)
        model.junk.setValue(model.z[1],4.0)
        
        self.assertTrue(model.junk.getValue(model.x) == -1.0)
        self.assertTrue(model.junk.getValue(model.y) == None)
        self.assertTrue(model.junk.getValue(model.y[1]) == -2.0)
        self.assertEqual(model.junk.getValue(model.y[2]), 1.0)
        self.assertEqual(model.junk.getValue(model.z), None)
        self.assertEqual(model.junk.getValue(model.z[2]), 2.0)
        self.assertEqual(model.junk.getValue(model.z[1]), 4.0)

        model.junk.clearValue(model.y)
        model.junk.clearValue(model.x)
        model.junk.clearValue(model.z[1])

        self.assertTrue(model.junk.getValue(model.x) is None)
        self.assertTrue(model.junk.getValue(model.y) is None)
        self.assertTrue(model.junk.getValue(model.y[1]) is None)
        self.assertEqual(model.junk.getValue(model.y[2]), None)
        self.assertEqual(model.junk.getValue(model.z), None)
        self.assertEqual(model.junk.getValue(model.z[2]), 2.0)
        self.assertEqual(model.junk.getValue(model.z[1]), None)

    # test clearValue no args
    def test_clearAllValues(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.x = Var()
        model.y = Var([1,2,3])
        model.z = Var([1,2,3])

        model.junk.setValue(model.y[2],1.0)
        model.junk.setValue(model.z,2.0)

        self.assertTrue(model.junk.getValue(model.x) is None)
        self.assertTrue(model.junk.getValue(model.y) is None)
        self.assertTrue(model.junk.getValue(model.y[1]) is None)
        self.assertEqual(model.junk.getValue(model.y[2]), 1.0)
        self.assertEqual(model.junk.getValue(model.z), None)
        self.assertEqual(model.junk.getValue(model.z[1]), 2.0)

        model.junk.clearAllValues()

        self.assertTrue(model.junk.getValue(model.x) is None)
        self.assertTrue(model.junk.getValue(model.y) is None)
        self.assertTrue(model.junk.getValue(model.y[1]) is None)
        self.assertTrue(model.junk.getValue(model.y[2]) is None)
        self.assertTrue(model.junk.getValue(model.z) is None)
        self.assertTrue(model.junk.getValue(model.z[1]) is None)

    # test setDatatype and getDatatype
    def test_setDatatype_getDatatype(self):
        model = ConcreteModel()
        model.junk = Suffix(datatype=Suffix.FLOAT)
        self.assertTrue(model.junk.getDatatype() is Suffix.FLOAT)
        model.junk.setDatatype(Suffix.INT)
        self.assertTrue(model.junk.getDatatype() is Suffix.INT)
        model.junk.setDatatype(None)
        self.assertTrue(model.junk.getDatatype() is None)

    # test that calling setDatatype with a bad value fails
    def test_setDatatype_badvalue(self):
        model = ConcreteModel()
        model.junk = Suffix()
        try:
            model.junk.setDatatype(1.0)
        except ValueError:
            pass
        else:
            self.fail("Calling setDatatype with a bad type should fail.")

    # test setDirection and getDirection
    def test_setDirection_getDirection(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.LOCAL)
        self.assertTrue(model.junk.getDirection() is Suffix.LOCAL)
        model.junk.setDirection(Suffix.EXPORT)
        self.assertTrue(model.junk.getDirection() is Suffix.EXPORT)
        model.junk.setDirection(Suffix.IMPORT)
        self.assertTrue(model.junk.getDirection() is Suffix.IMPORT)
        model.junk.setDirection(Suffix.IMPORT_EXPORT)
        self.assertTrue(model.junk.getDirection() is Suffix.IMPORT_EXPORT)
        
    # test that calling setDirection with a bad value fails
    def test_setDirection_badvalue(self):
        model = ConcreteModel()
        model.junk = Suffix()
        try:
            model.junk.setDirection('a')
        except ValueError:
            pass
        else:
            self.fail("Calling setDatatype with a bad type should fail.")    

    # test __str__
    def test_str(self):
        model = ConcreteModel()
        model.junk = Suffix()
        self.assertEqual(model.junk.__str__(), "junk")

    # test pprint()
    def test_pprint(self):
        model = ConcreteModel()
        model.junk = Suffix(direction=Suffix.EXPORT)
        output = StringIO()
        model.junk.pprint(ostream=output)
        model.junk.setDirection(Suffix.IMPORT)
        model.junk.pprint(ostream=output)
        model.junk.setDirection(Suffix.LOCAL)
        model.junk.pprint(ostream=output)
        model.junk.setDirection(Suffix.IMPORT_EXPORT)
        model.junk.pprint(ostream=output)
        model.pprint(ostream=output)

    # test pprint(verbose=True)
    def test_pprint_verbose(self):
        model = ConcreteModel()
        model.junk = Suffix()
        model.s = Block()
        model.s.b = Block()
        model.s.B = Block([1,2,3])

        model.junk.setValue(model.s.B,1.0)
        model.junk.setValue(model.s.B[1],2.0)

        model.junk.setValue(model.s.b,3.0)
        model.junk.setValue(model.s.B[2],3.0)

        output = StringIO()
        model.junk.pprint(ostream=output,verbose=True)
        model.pprint(ostream=output,verbose=True)

    def test_active_export_suffix_generator(self):
        model = ConcreteModel()
        model.junk_EXPORT_int = Suffix(direction=Suffix.EXPORT,datatype=Suffix.INT)
        model.junk_EXPORT_float = Suffix(direction=Suffix.EXPORT,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=Suffix.FLOAT)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT,datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL,datatype=None)

        suffixes = dict(active_export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(active_export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_EXPORT_float.activate()

        suffixes = dict(active_export_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(active_export_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_export_suffix_generator(self):
        model = ConcreteModel()
        model.junk_EXPORT_int = Suffix(direction=Suffix.EXPORT,datatype=Suffix.INT)
        model.junk_EXPORT_float = Suffix(direction=Suffix.EXPORT,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=Suffix.FLOAT)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT,datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL,datatype=None)

        suffixes = dict(export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(export_suffix_generator(model))
        self.assertTrue('junk_EXPORT_int' in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_EXPORT_float.activate()

        suffixes = dict(export_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_EXPORT_float.deactivate()
        suffixes = dict(export_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_EXPORT_int' not in suffixes)
        self.assertTrue('junk_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_active_import_suffix_generator(self):
        model = ConcreteModel()
        model.junk_IMPORT_int = Suffix(direction=Suffix.IMPORT,datatype=Suffix.INT)
        model.junk_IMPORT_float = Suffix(direction=Suffix.IMPORT,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=Suffix.FLOAT)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT,datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL,datatype=None)

        suffixes = dict(active_import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(active_import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_IMPORT_float.activate()

        suffixes = dict(active_import_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(active_import_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_import_suffix_generator(self):
        model = ConcreteModel()
        model.junk_IMPORT_int = Suffix(direction=Suffix.IMPORT,datatype=Suffix.INT)
        model.junk_IMPORT_float = Suffix(direction=Suffix.IMPORT,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT_float = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=Suffix.FLOAT)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT,datatype=None)
        model.junk_LOCAL = Suffix(direction=Suffix.LOCAL,datatype=None)

        suffixes = dict(import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(import_suffix_generator(model))
        self.assertTrue('junk_IMPORT_int' in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)
        model.junk_IMPORT_float.activate()

        suffixes = dict(import_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

        model.junk_IMPORT_float.deactivate()
        suffixes = dict(import_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_IMPORT_int' not in suffixes)
        self.assertTrue('junk_IMPORT_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT_float' in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_LOCAL' not in suffixes)

    def test_active_local_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL,datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT,datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT,datatype=None)

        suffixes = dict(active_local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(active_local_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_local_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

    def test_local_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL,datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT,datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT,datatype=None)

        suffixes = dict(local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(local_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(local_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(local_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

    def test_active_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL,datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT,datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT,datatype=None)

        suffixes = dict(active_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(active_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(active_suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' not in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

    def test_suffix_generator(self):
        model = ConcreteModel()
        model.junk_LOCAL_int = Suffix(direction=Suffix.LOCAL,datatype=Suffix.INT)
        model.junk_LOCAL_float = Suffix(direction=Suffix.LOCAL,datatype=Suffix.FLOAT)
        model.junk_IMPORT_EXPORT = Suffix(direction=Suffix.IMPORT_EXPORT,datatype=None)
        model.junk_EXPORT = Suffix(direction=Suffix.EXPORT,datatype=None)
        model.junk_IMPORT = Suffix(direction=Suffix.IMPORT,datatype=None)

        suffixes = dict(suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(suffix_generator(model))
        self.assertTrue('junk_LOCAL_int' in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' in suffixes)
        self.assertTrue('junk_EXPORT' in suffixes)
        self.assertTrue('junk_IMPORT' in suffixes)
        model.junk_LOCAL_float.activate()

        suffixes = dict(suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

        model.junk_LOCAL_float.deactivate()
        suffixes = dict(suffix_generator(model,datatype=Suffix.FLOAT))
        self.assertTrue('junk_LOCAL_int' not in suffixes)
        self.assertTrue('junk_LOCAL_float' in suffixes)
        self.assertTrue('junk_IMPORT_EXPORT' not in suffixes)
        self.assertTrue('junk_EXPORT' not in suffixes)
        self.assertTrue('junk_IMPORT' not in suffixes)

    def test_reset(self):
        model = ConcreteModel()
        model.x = Var()
        model.y = Var()
        model.junk_no_rule = Suffix()
        self.assertEqual(model.junk_no_rule.getValue(model.x),None)
        self.assertEqual(model.junk_no_rule.getValue(model.y),None)
        model.junk_no_rule.setValue(model.x,1)
        model.junk_no_rule.setValue(model.y,2)
        self.assertEqual(model.junk_no_rule.getValue(model.x),1)
        self.assertEqual(model.junk_no_rule.getValue(model.y),2)
        model.junk_no_rule.reset()
        self.assertEqual(model.junk_no_rule.getValue(model.x),None)
        self.assertEqual(model.junk_no_rule.getValue(model.y),None)
        
        model.del_component('junk_no_rule')

        def _junk_rule(model):
            return [(model.x,1)]
        model.junk_rule = Suffix(rule=_junk_rule)
        self.assertEqual(model.junk_rule.getValue(model.x),1)
        self.assertEqual(model.junk_rule.getValue(model.y),None)
        model.junk_rule.setValue(model.y,2)
        self.assertEqual(model.junk_rule.getValue(model.x),1)
        self.assertEqual(model.junk_rule.getValue(model.y),2)
        model.junk_rule.reset()
        self.assertEqual(model.junk_rule.getValue(model.x),1)
        self.assertEqual(model.junk_rule.getValue(model.y),None)

class TestSuffixCloneUsage(unittest.TestCase):

    def test_clone_VarElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.x),None)
        model.junk.setValue(model.x,1.0)
        self.assertEqual(model.junk.getValue(model.x),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.x),None)
        self.assertEqual(inst.junk.getValue(inst.x),1.0)

    def test_clone_VarArray(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.x),None)
        self.assertEqual(model.junk.getValue(model.x[1]),None)
        model.junk.setValue(model.x,1.0)
        self.assertEqual(model.junk.getValue(model.x),None)
        self.assertEqual(model.junk.getValue(model.x[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.x[1]),None)
        self.assertEqual(inst.junk.getValue(inst.x[1]),1.0)

    def test_clone_VarData(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.x[1]),None)
        model.junk.setValue(model.x[1],1.0)
        self.assertEqual(model.junk.getValue(model.x[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.x[1]),None)
        self.assertEqual(inst.junk.getValue(inst.x[1]),1.0)

    def test_clone_ConstraintElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=model.x == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.c),None)
        model.junk.setValue(model.c,1.0)
        self.assertEqual(model.junk.getValue(model.c),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.c),None)
        self.assertEqual(inst.junk.getValue(inst.c),1.0)

    def test_clone_ConstraintArray(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.c = Constraint([1,2,3],rule=lambda model,i: model.x[i] == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.c),None)
        self.assertEqual(model.junk.getValue(model.c[1]),None)
        model.junk.setValue(model.c,1.0)
        self.assertEqual(model.junk.getValue(model.c),None)
        self.assertEqual(model.junk.getValue(model.c[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.c[1]),None)
        self.assertEqual(inst.junk.getValue(inst.c[1]),1.0)

    def test_clone_ConstraintData(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.c = Constraint([1,2,3],rule=lambda model,i: model.x[i] == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.c[1]),None)
        model.junk.setValue(model.c[1],1.0)
        self.assertEqual(model.junk.getValue(model.c[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.c[1]),None)
        self.assertEqual(inst.junk.getValue(inst.c[1]),1.0)

    def test_clone_ObjectiveElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.obj = Objective(expr=model.x)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.obj),None)
        model.junk.setValue(model.obj,1.0)
        self.assertEqual(model.junk.getValue(model.obj),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.obj),None)
        self.assertEqual(inst.junk.getValue(inst.obj),1.0)

    def test_clone_ObjectiveArray(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.obj = Objective([1,2,3], rule=lambda model,i: model.x[i])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.obj),None)
        self.assertEqual(model.junk.getValue(model.obj[1]),None)
        model.junk.setValue(model.obj,1.0)
        self.assertEqual(model.junk.getValue(model.obj),None)
        self.assertEqual(model.junk.getValue(model.obj[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.obj[1]),None)
        self.assertEqual(inst.junk.getValue(inst.obj[1]),1.0)

    def test_clone_ObjectiveData(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.obj = Objective([1,2,3], rule=lambda model,i: model.x[i])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.obj[1]),None)
        model.junk.setValue(model.obj[1],1.0)
        self.assertEqual(model.junk.getValue(model.obj[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.obj[1]),None)
        self.assertEqual(inst.junk.getValue(inst.obj[1]),1.0)

    def test_clone_SimpleBlock(self):
        model = ConcreteModel()
        model.b = Block()
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.b),None)
        model.junk.setValue(model.b,1.0)
        self.assertEqual(model.junk.getValue(model.b),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.b),None)
        self.assertEqual(inst.junk.getValue(inst.b),1.0)

    def test_clone_IndexedBlock(self):
        model = ConcreteModel()
        model.b = Block([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.b),None)
        self.assertEqual(model.junk.getValue(model.b[1]),None)
        model.junk.setValue(model.b,1.0)
        self.assertEqual(model.junk.getValue(model.b),None)
        self.assertEqual(model.junk.getValue(model.b[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.b[1]),None)
        self.assertEqual(inst.junk.getValue(inst.b[1]),1.0)

    def test_clone_BlockData(self):
        model = ConcreteModel()
        model.b = Block([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.b[1]),None)
        model.junk.setValue(model.b[1],1.0)
        self.assertEqual(model.junk.getValue(model.b[1]),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model.b[1]),None)
        self.assertEqual(inst.junk.getValue(inst.b[1]),1.0)

    def test_clone_model(self):
        model = ConcreteModel()
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model),None)
        model.junk.setValue(model,1.0)
        self.assertEqual(model.junk.getValue(model),1.0)
        inst = model.clone()
        self.assertEqual(inst.junk.getValue(model),None)
        self.assertEqual(inst.junk.getValue(inst),1.0)

class TestSuffixPickleUsage(unittest.TestCase):

    def test_pickle_VarElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.x),None)
        model.junk.setValue(model.x,1.0)
        self.assertEqual(model.junk.getValue(model.x),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.x),None)
        self.assertEqual(inst.junk.getValue(inst.x),1.0)

    def test_pickle_VarArray(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.x),None)
        self.assertEqual(model.junk.getValue(model.x[1]),None)
        model.junk.setValue(model.x,1.0)
        self.assertEqual(model.junk.getValue(model.x),None)
        self.assertEqual(model.junk.getValue(model.x[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.x[1]),None)
        self.assertEqual(inst.junk.getValue(inst.x[1]),1.0)

    def test_pickle_VarData(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.x[1]),None)
        model.junk.setValue(model.x[1],1.0)
        self.assertEqual(model.junk.getValue(model.x[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.x[1]),None)
        self.assertEqual(inst.junk.getValue(inst.x[1]),1.0)

    def test_pickle_ConstraintElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.c = Constraint(expr=model.x == 1.0)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.c),None)
        model.junk.setValue(model.c,1.0)
        self.assertEqual(model.junk.getValue(model.c),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.c),None)
        self.assertEqual(inst.junk.getValue(inst.c),1.0)

    def test_pickle_ConstraintArray(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.c = Constraint([1,2,3],rule=simple_con_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.c),None)
        self.assertEqual(model.junk.getValue(model.c[1]),None)
        model.junk.setValue(model.c,1.0)
        self.assertEqual(model.junk.getValue(model.c),None)
        self.assertEqual(model.junk.getValue(model.c[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.c[1]),None)
        self.assertEqual(inst.junk.getValue(inst.c[1]),1.0)

    def test_pickle_ConstraintData(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.c = Constraint([1,2,3],rule=simple_con_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.c[1]),None)
        model.junk.setValue(model.c[1],1.0)
        self.assertEqual(model.junk.getValue(model.c[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.c[1]),None)
        self.assertEqual(inst.junk.getValue(inst.c[1]),1.0)

    def test_pickle_ObjectiveElement(self):
        model = ConcreteModel()
        model.x = Var()
        model.obj = Objective(expr=model.x)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.obj),None)
        model.junk.setValue(model.obj,1.0)
        self.assertEqual(model.junk.getValue(model.obj),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.obj),None)
        self.assertEqual(inst.junk.getValue(inst.obj),1.0)

    def test_pickle_ObjectiveArray(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.obj = Objective([1,2,3],rule=simple_obj_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.obj),None)
        self.assertEqual(model.junk.getValue(model.obj[1]),None)
        model.junk.setValue(model.obj,1.0)
        self.assertEqual(model.junk.getValue(model.obj),None)
        self.assertEqual(model.junk.getValue(model.obj[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.obj[1]),None)
        self.assertEqual(inst.junk.getValue(inst.obj[1]),1.0)

    def test_pickle_ObjectiveData(self):
        model = ConcreteModel()
        model.x = Var([1,2,3])
        model.obj = Objective([1,2,3],rule=simple_obj_rule)
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.obj[1]),None)
        model.junk.setValue(model.obj[1],1.0)
        self.assertEqual(model.junk.getValue(model.obj[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.obj[1]),None)
        self.assertEqual(inst.junk.getValue(inst.obj[1]),1.0)

    def test_pickle_SimpleBlock(self):
        model = ConcreteModel()
        model.b = Block()
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.b),None)
        model.junk.setValue(model.b,1.0)
        self.assertEqual(model.junk.getValue(model.b),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.b),None)
        self.assertEqual(inst.junk.getValue(inst.b),1.0)

    def test_pickle_IndexedBlock(self):
        model = ConcreteModel()
        model.b = Block([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.b),None)
        self.assertEqual(model.junk.getValue(model.b[1]),None)
        model.junk.setValue(model.b,1.0)
        self.assertEqual(model.junk.getValue(model.b),None)
        self.assertEqual(model.junk.getValue(model.b[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.b[1]),None)
        self.assertEqual(inst.junk.getValue(inst.b[1]),1.0)

    def test_pickle_BlockData(self):
        model = ConcreteModel()
        model.b = Block([1,2,3])
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model.b[1]),None)
        model.junk.setValue(model.b[1],1.0)
        self.assertEqual(model.junk.getValue(model.b[1]),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model.b[1]),None)
        self.assertEqual(inst.junk.getValue(inst.b[1]),1.0)

    def test_pickle_model(self):
        model = ConcreteModel()
        model.junk = Suffix()
        self.assertEqual(model.junk.getValue(model),None)
        model.junk.setValue(model,1.0)
        self.assertEqual(model.junk.getValue(model),1.0)
        inst = pickle.loads(pickle.dumps(model))
        self.assertEqual(inst.junk.getValue(model),None)
        self.assertEqual(inst.junk.getValue(inst),1.0)


if __name__ == "__main__":
    unittest.main()



from coopr.pyomo.base import Transformation


class AbstractTransformation(Transformation):
    """
    Base class for all model transformations that produce abstract
    models.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "abstract_transformation")
        super(AbstractTransformation, self).__init__(**kwds)


class ConcreteTransformation(Transformation):
    """
    Base class for all model transformations that produce concrete
    models.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "concrete_transformation")
        super(ConcreteTransformation, self).__init__(**kwds)


class IsomorphicTransformation(Transformation):
    """
    Base class for 'lossless' transformations for which a bijective
    mapping between optimal variable values and the optimal cost
    exists.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "isomorphic_transformation")
        super(IsomorphicTransformation, self).__init__(**kwds)


class LinearTransformation(Transformation):
    """ Base class for all linear model transformations. """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "linear_transform")
        super(LinearTransformation, self).__init__(**kwds)


class NonIsomorphicTransformation(Transformation):
    """
    Base class for 'lossy' transformations for which a bijective
    mapping between optimal variable values and the optimal cost does
    not  exist.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "isomorphic_transformation")
        super(NonIsomorphicTransformation, self).__init__(**kwds)


class NonlinearTransformation(Transformation):
    """ Base class for all nonlinear model transformations. """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "nonlinear_transform")
        super(NonlinearTransformation, self).__init__(**kwds)




import os
import tempfile

import pyutilib.th as unittest
from nose.tools import nottest
from coopr.pyomo.check import *

currdir = os.path.dirname(os.path.abspath(__file__))

class ScriptTest(unittest.TestCase):
    """
    Test the ModelScript class. Checks both raw text and file-based
    interfaces (using the tempfile module).
    """

    testScripts = [
        "print('Hello, world!')\n",
        "import sys\nsys.stdout.write('Hello, world!\\n')\n"
        "for i in range(10):\n\tprint(i)\n"
    ]

    def testScriptText(self):
        "Check ModelScript handling of raw text scripts"

        for text in self.testScripts:
            script = ModelScript(text = text)
            self.assertEqual(text, script.read())
            self.assertEqual("<unknown>", script.filename())

    def testScriptFile(self):
        "Check ModelScript handling of file-based scripts"

        for text in self.testScripts:
            file, filename = tempfile.mkstemp()

            with os.fdopen(file, 'w') as f:
                f.write(text)

            script = ModelScript(filename = filename)

            self.assertEqual(text, script.read())
            self.assertEqual(filename, script.filename())

            os.unlink(filename)


if __name__ == "__main__":
    unittest.main()




from coopr.core.plugin import *


class IModelChecker(Interface):

    def check(self, runner, script, info):
        """
        Check a particular piece of Python information for errors.

        Provides the primary interface for checking some code for problems,
        according to the particular subclass's definition. 
        
        @param runner the ModelCheckRunner instance that has dispatched
                      this call to check().
        @param script the ModelScript instance being checked.
        @param info the data to check. Depending on the subclass, info
                    can be the raw text of a Python script, the entire AST
                    of the script, or a particular node in that AST.
        """

    def beginChecking(self, runner, script):
        """
        Start checking the given script from the given runner.
        """

    def endChecking(self, runner, script):
        """
        Finish checking the given script from the given runner.
        """

    def problem(self, script = None, message = "Error", lineno = None):
        """
        Write a problem to the console. The format varies and can be
        changed in subclasses; by default, this method prints the following:

        [CheckerName] script.py:line: Error

        @param script the ModelScript instance being checked. Must be passed
                      to have the file name and line number printed.
        @param message the error to display.
        @param lineno the line number on which the error occurred.
        """




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import sys
import re
import copy
import math
import logging

from six.moves import xrange

from pyutilib.misc import quote_split, Options
import pyutilib.common
from pyutilib.misc import flatten

from coopr.pyomo.base.plugin import *
from coopr.pyomo.base.plugin import *
from coopr.pyomo.base.param import Param
from coopr.pyomo.base.sets import Set
from coopr.pyomo.data.parse_datacmds import parse_data_commands

try:
    unicode
except:
    unicode = str
try:
    long
    numlist = (bool, int, float, long)
except:
    numlist = (bool, int, float)


logger = logging.getLogger('coopr.pyomo')

global Lineno
global Filename


def _preprocess_data(cmd):
    """
    Called by _process_data() to (1) combine tokens that comprise a tuple
    and (2) combine the ':' token with the previous token
    """
    if __debug__:
        logger.debug("_preprocess_data(start) %s",cmd)
    status=")"
    newcmd=[]
    for token in cmd:
        if type(token) in (str,unicode):
            token=str(token)
            if "(" in token and ")" in token:
                newcmd.append(token)
                status=")"
            elif "(" in token:
                if status == "(":
                    raise ValueError("Two '('s follow each other in data "+token)
                status="("
                newcmd.append(token)
            elif ")" in token:
                if status == ")":
                    raise ValueError("Two ')'s follow each other in data")
                status=")"
                newcmd[-1] = newcmd[-1]+token
            elif status == "(":
                newcmd[-1] = newcmd[-1]+token
            else:
                newcmd.append(token)
        else:
            if type(token) is float and math.floor(token) == token:
                token=int(token)
            newcmd.append(token)
    if __debug__:
        logger.debug("_preprocess_data(end) %s", newcmd)
    return newcmd


def _process_set(cmd, _model, _data):
    """
    Called by _process_data() to process a set declaration.
    """
    #print "SET",cmd
    if __debug__:
        logger.debug("DEBUG: _process_set(start) %s",cmd)
    #
    # Process a set
    #
    if "[" in cmd[1]: 
        # the set is indexed
        tokens = re.split("[\[\]]",cmd[1])
        ndx=tokens[1]
        if len(ndx) == 0:
            # at this point, the index is a string. if the length of this string
            # is 0, then there is a specification issue, as this is an indexed set.
            raise ValueError("Illegal indexed set specification encountered: "+str(cmd[1]))
        ndx=tuple(_data_eval(ndx.split(",")))
        if tokens[0] not in _data:
            _data[tokens[0]] = {}
        _data[tokens[0]][ndx] = _process_set_data(cmd[3:], tokens[0], _model)
    elif cmd[2] == ":":
        _data[cmd[1]] = {}
        _data[cmd[1]][None] = []
        i=3
        while cmd[i] != ":=":
            i += 1
        ndx1 = cmd[3:i]
        i += 1
        while i<len(cmd):
            ndx=cmd[i]
            for j in xrange(0,len(ndx1)):
                if cmd[i+j+1] == "+":
                    _data[cmd[1]][None] += _process_set_data(["("+str(ndx1[j])+","+str(cmd[i])+")"], cmd[1], _model)
            i += len(ndx1)+1
    else:
        _data[cmd[1]] = {}
        _data[cmd[1]][None] = _process_set_data(cmd[3:], cmd[1], _model)


def _process_set_data(cmd, sname, _model):
    """
    Called by _process_set() to process set data.
    """
    if __debug__:
        logger.debug("DEBUG: _process_set_data(start) %s",cmd)
    if len(cmd) == 0:
        return []
    sd = sname
    #d = sd.dimen
    cmd = _data_eval(cmd)
    ans=[]
    i=0
    tmp=None
    ndx=[]
    while i<len(cmd):
        if type(cmd[i]) is not tuple:
            if len(ndx) > 0:
                #if type(cmd[i]) is not tuple:
                #   raise ValueError, "Problem initializing set="+sname+" with input data="+str(cmd)+" - first element was interpreted as a tuple, but element="+str(i)+" is of type="+str(type(cmd[i]))+"; types must be consistent"
                tmpval=tmp
                for kk in range(len(ndx)):
                    if i == len(cmd):
                        raise IOError("Expected another set value to flush out a tuple pattern!")
                    tmpval[ndx[kk]] = _data_eval([cmd[i]])[0]
                    i += 1
                #
                # WEH - I'm not sure what the next two lines are for
                #        These are called when initializing a set with more than
                #        one dimension
                #if d > 1:
                #   tmpval = util.tuplize(tmpval,d,sname)
                ans.append(tuple(tmpval))
                continue
            else:
                ans.append(cmd[i])
        elif "*" not in cmd[i]:
            ans.append(cmd[i])
        else:
            j = i
            tmp=list(cmd[j])
            ndx=[]
            for kk in range(len(tmp)):
                if tmp[kk] == '*':
                    ndx.append(kk)
            #print 'NDX',ndx
        i += 1
    if __debug__:
        logger.debug("DEBUG: _process_set_data(end) %s",ans)
    return ans


def _process_param(cmd, _model, _data, _default, index=None, param=None, ncolumns=None):
    """
    Called by _process_data to process data for a Parameter declaration
    """
    #print 'PARAM',cmd,index, ncolumns
    if __debug__:
        logger.debug("DEBUG: _process_param(start) %s",cmd)
    #
    # Process parameters
    #
    dflt = None
    singledef = True
    cmd = cmd[1:]
    if cmd[0] == ":":
        singledef = False
        cmd = cmd[1:]
    #print "SINGLEDEF", singledef
    if singledef:
        pname = cmd[0]
        cmd = cmd[1:]
        if len(cmd) >= 2 and cmd[0] == "default":
            dflt = _data_eval(cmd[1])[0]
            cmd = cmd[2:]
        if dflt != None:
            _default[pname] = dflt
        if cmd[0] == ":=":
            cmd = cmd[1:]
        transpose = False
        if cmd[0] == "(tr)":
            transpose = True
            if cmd[1] == ":":
                cmd = cmd[1:]
            else:
                cmd[0] = ":"
        if cmd[0] != ":":
            #print "HERE YYY", pname, transpose, _model, ncolumns
            if __debug__:
                logger.debug("DEBUG: _process_param (singledef without :...:=) %s",cmd)
            cmd = _apply_templates(cmd)
            #print 'cmd',cmd
            if not transpose:
                if pname not in _data:
                    _data[pname] = {}
                if not ncolumns is None:
                    finaldata = _process_data_list(pname, ncolumns-1, _data_eval(cmd))
                elif not _model is None:
                    _param = getattr(_model, pname)
                    finaldata = _process_data_list(pname, _param.dim(), _data_eval(cmd))
                else:
                    finaldata = _process_data_list(pname, 1, _data_eval(cmd))
                for key in finaldata:
                    _data[pname][key]=finaldata[key]
            else:
                tmp = ["param", pname, ":="]
                i=1
                while i < len(cmd):
                    i0 = i
                    while cmd[i] != ":=":
                        i=i+1
                    ncol = i - i0 + 1
                    lcmd = i
                    while lcmd < len(cmd) and cmd[lcmd] != ":":
                        lcmd += 1
                    j0 = i0 - 1
                    for j in range(1,ncol):
                        ii = 1 + i
                        kk = ii + j
                        while kk < lcmd:
                            if cmd[kk] != ".":
                            #if 1>0:
                                tmp.append(copy.copy(cmd[j+j0]))
                                tmp.append(copy.copy(cmd[ii]))
                                tmp.append(copy.copy(cmd[kk]))
                            ii = ii + ncol
                            kk = kk + ncol
                    i = lcmd + 1
                _process_param(tmp, _model, _data, _default, index=index, param=param, ncolumns=ncolumns)
        else:
            tmp = ["param", pname, ":="]
            if param is None:
                param = [ pname ]
            i=1
            if __debug__:
                logger.debug("DEBUG: _process_param (singledef with :...:=) %s",cmd)
            while i < len(cmd):
                i0 = i
                while i<len(cmd) and cmd[i] != ":=":
                    i=i+1
                if i==len(cmd):
                    raise ValueError("ERROR: Trouble on line "+str(Lineno)+" of file "+Filename)
                ncol = i - i0 + 1
                lcmd = i
                while lcmd < len(cmd) and cmd[lcmd] != ":":
                    lcmd += 1
                j0 = i0 - 1
                for j in range(1,ncol):
                    ii = 1 + i
                    kk = ii + j
                    while kk < lcmd:
                        if cmd[kk] != ".":
                            if transpose:
                                tmp.append(copy.copy(cmd[j+j0]))
                                tmp.append(copy.copy(cmd[ii]))
                            else:
                                tmp.append(copy.copy(cmd[ii]))
                                tmp.append(copy.copy(cmd[j+j0]))
                            tmp.append(copy.copy(cmd[kk]))
                        ii = ii + ncol
                        kk = kk + ncol
                i = lcmd + 1
                _process_param(tmp, _model, _data, _default, index=index, param=param[0], ncolumns=3)

    else:
        if __debug__:
            logger.debug("DEBUG: _process_param (cmd[0]=='param:') %s",cmd)
        i=0
        nsets=0
        while i<len(cmd):
            if cmd[i] == ':=':
                i = -1
                break
            if cmd[i] == ":":
                nsets = i
                break
            i += 1
        nparams=0
        _i = i+1
        while i<len(cmd):
            if cmd[i] == ':=':
                nparams = i-_i
                break
            i += 1
        if i==len(cmd):
            raise ValueError("Trouble on data file line "+str(Lineno)+" of file "+Filename)
        if __debug__:
            logger.debug("NSets %d",nsets)
        Lcmd = len(cmd)
        j=0
        d = 1
        #print "HERE", nsets, nparams
        #
        # Process sets first
        #
        while j<nsets:
            # NOTE: I'm pretty sure that nsets is always equal to 1
            sname = cmd[j]
            if not ncolumns is None:
                d = ncolumns-nparams
            elif _model is None:
                d = 1
            else:
                index = getattr(_model, sname)
                d = index.dimen
            #print "SET",sname,d,_model#,getattr(_model,sname).dimen, type(index)
            #d = getattr(_model,sname).dimen
            np = i-1
            if __debug__:
                logger.debug("I %d, J %d, SName %s, d %d",i,j,sname,d)
            dnp = d + np - 1
            #k0 = i + d - 2
            ii = i + j + 1
            tmp = [ "set", cmd[j], ":=" ]
            while ii < Lcmd:
                if d > 1:
                    _tmp = []
                    for dd in range(0,d):
                        _tmp.append(copy.copy(cmd[ii+dd]))
                    tmp.append(tuple(_tmp))
                else:
                    for dd in range(0,d):
                        tmp.append(copy.copy(cmd[ii+dd]))
                ii += dnp
            _process_set(tmp, _model, _data)
            j += 1
        if nsets > 0:
            j += 1
        #
        # Process parameters second
        #
        #print "HERE", cmd
        #print "HERE", param
        #print "JI",j,i # XXX
        jstart = j
        if param is None:
            param = []
            _j = j
            while _j < i:
                param.append( cmd[_j] )
                _j += 1
        while j < i:
            #print "HERE", i, j, jstart, cmd[j]
            pname = param[j-jstart]
            if __debug__:
                logger.debug("I %d, J %d, Pname %s",i,j,pname)
            if not ncolumns is None:
                d = ncolumns - nparams
            elif _model is None:
                d = 1
            else:
                d = getattr(_model, param[j-jstart]).dim()
            if nsets > 0:
                np = i-1
                dnp = d+np-1
                ii = i + 1
                kk = i + d + j-1
            else:
                np = i
                dnp = d + np
                ii = i + 1
                kk = np + 1 + d + nsets + j
            #print cmd[ii], d, np, dnp, ii, kk
            tmp = [ "param", pname, ":=" ]
            if __debug__:
                logger.debug('dnp %d\nnp %d', dnp, np)
            while kk < Lcmd:
                if __debug__:
                    logger.debug("kk %d, ii %d",kk,ii)
                iid = ii + d
                while ii < iid:
                    tmp.append(copy.copy(cmd[ii]))
                    ii += 1
                ii += dnp-d
                tmp.append(copy.copy(cmd[kk]))
                kk += dnp
            #print "TMP", tmp, ncolumns-nparams+1
            if not ncolumns is None:
                nc = ncolumns-nparams+1
            else: 
                nc = None
            _process_param(tmp, _model, _data, _default, index=index, param=param[j-jstart], ncolumns=nc)
            j += 1


def _apply_templates(cmd):
    cmd = _data_eval(cmd)
    template = []
    ilist = set()
    ans = []
    i = 0
    while i < len(cmd):
    #print i, len(cmd), cmd[i], ilist, template, ans
        if type(cmd[i]) is tuple and '*' in cmd[i]:
            j = i
            tmp=list(cmd[j])
            nindex = len(tmp)
            template=tmp
            ilist = set()
            for kk in range(len(tmp)):
                if tmp[kk] == '*':
                    ilist.add(kk)
        elif len(ilist) == 0:
            ans.append(cmd[i])
        else:
            for kk in range(len(template)):
                if kk in ilist:
                    ans.append(cmd[i])
                    i += 1
                else:
                    ans.append(template[kk])
            ans.append(cmd[i])
        i += 1
    return ans


def _process_data_list(param_name, dim, cmd):
    """\
 Called by _process_param() to process a list of data for a Parameter.
 """
    if __debug__:
        logger.debug("process_data_list %d %s",dim,cmd)

    if len(cmd) % (dim+1) != 0:
        msg = "Parameter '%s' defined with '%d' dimensions, " \
              "but data has '%d' values: %s."
        msg = msg % (param_name, dim, len(cmd), cmd)
        if len(cmd) % (dim+1) == dim:
            msg += " Are you missing a value for a %d-dimensional index?" % dim
        elif len(cmd) % dim == 0:
            msg += " Are you missing the values for %d-dimensional indices?" % dim
        else:
            msg += " Data declaration must be given in multiples of %d." % (dim + 1)
        raise ValueError(msg)

    ans={}
    if dim == 0:
        ans[None]=cmd[0]
        return ans
    i=0
    while i < len(cmd):
        if dim > 1:
            ndx = tuple(cmd[i:i+dim])
        else:
            ndx = cmd[i]
        if cmd[i+dim] != ".":
            ans[ndx] = cmd[i+dim]
        i += dim+1
    return ans



def _data_eval(values):
    """
    Evaluate the list of values to make them bool, integer or float,
    or a tuple value.
    """
    if __debug__:
        logger.debug("DEBUG: _data_eval(start) %s",values)
    ans = []
    for val in values:
        if type(val) in numlist:
            ans.append(val)
            continue
        if val in ('True','true','TRUE'):
            ans.append(True)
            continue
        if val in ('False','false','FALSE'):
            ans.append(False)
            continue
        if type(val) is tuple:
            vals = []
            for item in val:
                vals.append( _data_eval([item])[0] )
            ans.append(tuple(vals))
            continue
        tmp = None
        tval = val.strip()
        if (tval[0] == "(" and tval[-1] == ")") or (tval[0] == '[' and tval[-1] == ']'):
            vals = []
            tval = tval[1:-1]
            for item in tval.split(","):
                tmp=_data_eval([item])
                vals.append(tmp[0])
            ans.append(tuple(vals))
            continue
        try:
            tmp = int(val)
            ans.append(tmp)
        except ValueError:
            pass
        if tmp is None:
            try:
                tmp = float(val)
                ans.append(tmp)
            except ValueError:
                if val[0] == "'" or val[0] == '"':
                    ans.append(val[1:-1])
                else:
                    ans.append(val)
    if __debug__:
        logger.debug("DEBUG: _data_eval(end) %s",ans)
    return ans


def _process_include(cmd, _model, _data, _default, options=None):
    if len(cmd) == 1:
        raise IOError("Cannot execute 'include' command without a filename")
    if len(cmd) > 2:
        raise IOError("The 'include' command only accepts a single filename")

    global Filename
    Filename = cmd[1]
    global Lineno
    Lineno = 0

    try:
        scenarios = parse_data_commands(filename=cmd[1])
    except IOError:
        err = sys.exc_info()[1]
        raise IOError("Error parsing file '%s': %s" % (Filename, str(err)))
    if scenarios is None:
        return False
    for scenario in scenarios:
        for cmd in scenarios[scenario]:
            if scenario not in _data:
                _data[scenario] = {}
            if cmd[0] in ('include', 'import', 'load'):
                _tmpdata = {}
                _process_data(cmd, _model, _tmpdata, _default, Filename, Lineno)
                if scenario is None:
                    for key in _tmpdata:
                        if key in _data:
                            _data[key].update(_tmpdata[key])
                        else:
                            _data[key] = _tmpdata[key]
                else:
                    for key in _tmpdata:
                        if key is None:
                            _data[scenario].update(_tmpdata[key])
                        else:
                            raise IOError("Cannot define a scenario within another scenario")
            else:
                _process_data(cmd, _model, _data[scenario], _default, Filename, Lineno)
    return True


def X_process_include(cmd, _model, _data, _default):
    if len(cmd) == 1:
        raise IOError("Cannot execute 'include' command without a filename")
    if len(cmd) > 2:
        raise IOError("The 'include' command only accepts a single filename")

    global Filename
    Filename = cmd[1]
    global Lineno
    Lineno = 0
    cmd=""
    status=True
    INPUT=open(Filename,'r')
    for line in INPUT:
        Lineno = Lineno + 1
        line = re.sub(":"," :",line)
        line = line.strip()
        if line == "" or line[0] == '#':
            continue
        cmd = cmd + " " + line
        if ';' in cmd:
            #
            # We assume that a ';' indicates an end-of-command declaration.
            # However, the user might have put multiple commands on a single
            # line, so we need to split the line based on these values.
            # BUT, at the end of the line we should see an 'empty' command,
            # which we ignore.
            #
            for item in cmd.split(';'):
                item = item.strip()
                if item != "":
                    _process_data(quote_split("[\t ]+",item), _model, _data, _default, Filename, Lineno)
                cmd = ""
    if cmd != "":
        INPUT.close()
        raise IOError("ERROR: There was unprocessed text at the end of the data file!: \"" + cmd + "\"")
    INPUT.close()
    return status


def _process_table(cmd, _model, _data, _default, options=None):
    #print "TABLE", cmd
    #
    options = Options(**cmd[1])
    for key in options:
        if not key in ['columns']:
            raise ValueError("Unknown table option '%s'" % key)
    #
    ncolumns=options.columns
    if ncolumns is None:
        ncolumns = len(cmd[4])
        if ncolumns == 0:
            if not (len(cmd[3]) == 1 and len(cmd[3][cmd[3].keys()[0]]) == 0):
                raise IOError("Must specify either the 'columns' option or column headers")
            else:
                ncolumns=1
    else:
        ncolumns = int(ncolumns)
    #
    data = cmd[5]
    Ldata = len(cmd[5])
    #
    cmap = {}
    if len(cmd[4]) == 0:
        for i in range(ncolumns):
            cmap[str(i+1)] = i
        for label in cmd[3]:
            ndx = int(cmd[3][label][1])-1
            if ndx < 0 or ndx >= ncolumns:
                raise IOError("Bad column value %s for data %s" % (str(ndx), label))
            cmap[label] = ndx
            cmd[3][label] = cmd[3][label][0]
    else:
        i = 0
        for label in cmd[4]:
            cmap[label] = i
            i += 1
    #
    for sname in cmd[2]:
        # Creating set sname
        cols = cmd[2][sname]
        tmp = []
        for col in cols:
            if not col in cmap:
                raise IOError("Unexpected table column '%s' for index set '%s'" % (col, sname))
            tmp.append(cmap[col])
        cmap[sname] = tmp
        cols = tmp
        # 
        _cmd = ['set', sname, ':=']
        i = 0
        while i < Ldata:
            row = []
            for col in cols:
                row.append( data[i+col] )
            if len(row) > 1:
                    _cmd.append( tuple(row) )
            else:
                    _cmd.append( row[0] )
            i += ncolumns
        _process_set(_cmd, _model, _data)
    #
    _i=0
    #print "HERE ncol", ncolumns
    if ncolumns == 0:
        raise IOError
    for vname in cmd[3]:
        _i += 1
        # create value vname
        cols = cmd[3][vname]
        tmp = []
        for col in cols:
            if not col in cmap:
                raise IOError("Unexpected table column '%s' for table value '%s'" % (col, vname))
            tmp.append(cmap[col])
        cols = flatten(tmp)
        if vname in cmap:
            cols.append(cmap[vname])
        else:
            cols.append( ncolumns-1 - (len(cmd[3])-_i) )
        #
        _cmd = ['param', vname, ':=']
        i = 0
        while i < Ldata:
            #print "HERE", i, cols, ncolumns
            for col in cols:
                _cmd.append( data[i+col] )
            i += ncolumns
        _process_param(_cmd, _model, _data, None, ncolumns=len(cols))
            

def _process_import(cmd, _model, _data, _default, options=None):
    #logger.warning("WARNING: the 'import' data command is deprecated")

    #print "IMPORT",cmd
    if len(cmd) < 2:
        raise IOError("The 'import' command must specify a filename")

    options = Options(**cmd[1])
    for key in options:
        if not key in ['range','filename','format','using','driver','query','table','user','password']:
            raise ValueError("Unknown import option '%s'" % key)

    global Filename
    Filename = cmd[1]
    global Lineno
    Lineno = 0

    #
    # TODO: process mapping info
    #
    if options.using is None:
        tmp = options.filename.split(".")[-1]
        data = DataManagerFactory(tmp)
        if data is None:
            raise pyutilib.common.ApplicationError("Data manager '%s' is not available." % tmp)
    else:
        data = DataManagerFactory(options.using)
        if data is None:
            raise pyutilib.common.ApplicationError("Data manager '%s' is not available." % options.using)
    set_name=None
    param_name=None
    #
    # Create symbol map
    #
    symb_map = cmd[3]
    if len(symb_map) == 0:
        raise IOError("Must specify at least one set or parameter name that will be imported")
    #
    # Process index data
    #
    _index=None
    index_name=cmd[2][0]
    _select = None
    #
    # Set the 'set name' based on the format
    #
    _set = None
    if options.format == 'set' or options.format == 'set_array':
        if len(cmd[3]) != 1:
            raise IOError("A single set name must be specified when using format '%s'" % options.format)
        set_name=list(cmd[3].keys())[0]
        _set = set_name
    #
    # Set the 'param name' based on the format
    #
    _param = None
    if options.format == 'transposed_array' or options.format == 'array' or options.format == 'param':
        if len(cmd[3]) != 1:
            raise IOError("A single parameter name must be specified when using format '%s'" % options.format)
    if options.format in ('transposed_array', 'array', 'param', None):
        if cmd[2][0] is None:
            _index = None
        else:
            _index = cmd[2][0]
        _param = []
        _select = cmd[2][1]
        for key in cmd[3]:
            _param.append( cmd[3][key] )
            _select.append( key )
    if options.format in ('transposed_array', 'array'):
        _select = None

    #print "YYY", _param, options
    if not _param is None and len(_param) == 1 and not _model is None and isinstance(getattr(_model, _param[0]), Set):
        _select = None
        _set = _param[0]
        _param = None
        _index = None

    #print "SELECT", _param, _select
    #
    data.initialize(model=options.model, filename=options.filename, index=_index, index_name=index_name, param_name=symb_map, set=_set, param=_param, format=options.format, range=options.range, query=options.query, using=options.using, table=options.table, select=_select)
    #
    data.open()
    try:
        data.read()
    except Exception:
        data.close()
        raise
    data.close()
    data.process(_model, _data, _default)


def _process_load(cmd, _model, _data, _default, options=None):
    #logger.warning("WARNING: the 'import' data command is deprecated")

    #print "LOAD",cmd
    options = Options(**cmd[1])
    for key in options:
        if not key in ['range','filename','format','using','driver','query','table','user','password']:
            raise ValueError("Unknown import option '%s'" % key)

    global Filename
    Filename = cmd[4]
    global Lineno
    Lineno = 0

    #
    # TODO: process mapping info
    #
    if options.using is None:
        tmp = options.filename.split(".")[-1]
        data = DataManagerFactory(tmp)
        if data is None:
            raise pyutilib.common.ApplicationError("Data manager '%s' is not available." % tmp)
    else:
        data = DataManagerFactory(options.using)
        if data is None:
            raise pyutilib.common.ApplicationError("Data manager '%s' is not available." % options.using)
    set_name=None
    param_name=None
    #
    # Create symbol map
    #
    symb_map = cmd[3]
    if len(symb_map) == 0:
        raise IOError("Must specify at least one set or parameter name that will be imported")
    #
    # Process index data
    #
    _index=None
    index_name=cmd[2][0]
    _select = None
    #
    # Set the 'set name' based on the format
    #
    _set = None
    if options.format == 'set' or options.format == 'set_array':
        if len(cmd[3]) != 1:
            raise IOError("A single set name must be specified when using format '%s'" % options.format)
        set_name=cmd[3].keys()[0]
        _set = set_name
    #
    # Set the 'param name' based on the format
    #
    _param = None
    if options.format == 'transposed_array' or options.format == 'array' or options.format == 'param':
        if len(cmd[3]) != 1:
            raise IOError("A single parameter name must be specified when using format '%s'" % options.format)
    if options.format in ('transposed_array', 'array', 'param', None):
        if cmd[2][0] is None:
            _index = None
        else:
            _index = cmd[2][0]
        _param = []
        _select = cmd[2][1]
        for key in cmd[3]:
            _param.append( cmd[3][key] )
            _select.append( key )
    if options.format in ('transposed_array', 'array'):
        _select = None

    #print "YYY", _param, options
    if not _param is None and len(_param) == 1 and not _model is None and isinstance(getattr(_model, _param[0]), Set):
        _select = None
        _set = _param[0]
        _param = None
        _index = None

    #print "SELECT", _param, _select
    #
    data.initialize(model=options.model, filename=options.filename, index=_index, index_name=index_name, param_name=symb_map, set=_set, param=_param, format=options.format, range=options.range, query=options.query, using=options.using, table=options.table, select=_select)
    #
    data.open()
    try:
        data.read()
    except Exception:
        data.close()
        raise
    data.close()
    data.process(_model, _data, _default)


def _process_data(cmd, _model, _data, _default, Filename_, Lineno_=0, index=None, set=None, param=None, ncolumns=None):
    """
    Called by import_file() to (1) preprocess data and (2) call
    subroutines to process different types of data
    """
    #print "CMD",cmd
    global Lineno
    global Filename
    Lineno=Lineno_
    Filename=Filename_

    if __debug__:
        logger.debug("DEBUG: _process_data (start) %s",cmd)
    if len(cmd) == 0:                       #pragma:nocover
        raise ValueError("ERROR: Empty list passed to Model::_process_data")

    cmd = _preprocess_data(cmd)

    if cmd[0] == "data":
        return True
    if cmd[0] == "end":
        return False
    if cmd[0].startswith('set'):
        _process_set(cmd, _model, _data)
    elif cmd[0].startswith('param'):
        _process_param(cmd, _model, _data, _default, index=index, param=param, ncolumns=ncolumns)
    elif cmd[0] == 'include':
        _process_include(cmd, _model, _data, _default)
    elif cmd[0] == 'import':
        _process_import(cmd, _model, _data, _default)
    elif cmd[0] == 'load':
        _process_load(cmd, _model, _data, _default)
    elif cmd[0] == 'table':
        _process_table(cmd, _model, _data, _default)
    else:
        raise IOError("ERROR: Unknown data command: "+" ".join(cmd))
    return True



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['TableData']

from six.moves import xrange

from coopr.core.plugin import Plugin, implements
from pyutilib.misc import Options

from coopr.pyomo.base import value
from coopr.pyomo.base.plugin import IDataManager
from coopr.pyomo.base.sets import Set
from coopr.pyomo.base.param import Param
from coopr.pyomo.data.process_data import _process_data


class TableData(Plugin):
    """
    An object that imports data from a table in an external data source.
    """

    implements(IDataManager, service=False)

    def __init__(self):
        """
        Constructor
        """
        self._info=None
        self._data=None
        self.options = Options()
        self.options.ncolumns = 1

    def initialize(self, **kwds):
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        self.options.update(kwds)

    def open(self):                        #pragma:nocover
        """
        Open the table
        """
        pass

    def read(self):                         #pragma:nocover
        """
        Read data from the table
        """
        return False

    def write(self, data):                  #pragma:nocover
        """
        Write data from the table
        """
        return False

    def close(self):                        #pragma:nocover
        """
        Close the table
        """
        pass

    def process(self, model, data, default):
        """
        Return the data that was extracted from this table
        """
        if model is None:
            model = self.options.model
        if not None in data:
            data[None] = {}
        return _process_data(
          self._info,
          model,
          data[None],
          default,
          self.filename,
          index=self.options.index,
          set=self.options.set,
          param=self.options.param,
          ncolumns = self.options.ncolumns
        )

    def clear(self):
        """
        Clear the data that was extracted from this table
        """
        self._info = None

    def _set_data(self, headers, rows):
        header_index = []
        if self.options.select is None:
            for i in xrange(len(headers)):
                header_index.append(i)
        else:
            for i in headers:
                if str(i) in self.options.select:
                    header_index.append( self.options.select.index(str(i)) )
        self.options.ncolumns = len(headers)
        
        if not self.options.param is None:
            if not type(self.options.param) in (list, tuple):
                self.options.param = (self.options.param, )
            _params = []
            for p in self.options.param:
                if isinstance(p, Param):
                    self.options.model = p.model()
                    _params.append( p.name )
                else:
                    _params.append( p )
            self.options.param = tuple(_params)

        if isinstance(self.options.set, Set):
            self.options.model = self.options.set.model()
            self.options.set = self.options.set.name

        if isinstance(self.options.index, Set):
            self.options.model = self.options.index.model()
            self.options.index = self.options.index.name

        if self.options.format is None:
            if not self.options.set is None:
                self.options.format = 'set'
            elif not self.options.param is None:
                self.options.format = 'table'
            if self.options.format is None:
                raise ValueError("Unspecified format and  data option")
        elif self.options.set is None and self.options.param is None:
            msg = "Must specify the set or parameter option for data"
            raise IOError(msg)

        if self.options.format == 'set':
            if not self.options.index is None:
                msg = "Cannot specify index for data with the 'set' format: %s"
                raise IOError(msg % str( self.options.index ))

            self._info = ["set",self.options.set,":="]
            for row in rows:
                if self.options.ncolumns > 1:
                    self._info.append(tuple(row))
                else:
                    self._info.extend(row)

        elif self.options.format == 'set_array':
            if not self.options.index is None:
                msg = "Cannot specify index for data with the 'set_array' "   \
                      'format: %s'
                raise IOError(msg % str( self.options.index ))

            self._info = ["set",self.options.set, ":"]
            self._info = self._info + list(headers[1:])
            self._info = self._info + [":="]
            for row in rows:
                self._info = self._info + list(row)

        elif self.options.format == 'transposed_array':
            self._info = ["param",self.options.param[0],"(tr)",":"]              \
                       + headers[1:]
            self._info.append(":=")
            for row in rows:
                self._info.extend(row)

        elif self.options.format == 'array':
            self._info = ["param",self.options.param[0],":"] + headers[1:]
            self._info.append(":=")
            for row in rows:
                self._info.extend(row)

        elif self.options.format == 'table':
            if self.options.index is not None:
                self._info = ["param",":",self.options.index,":"]
            else:
                self._info = ["param",":"]
            for param in self.options.param:
                self._info.append(param)
            self._info.append(":=")
            for row in rows:
                for i in header_index:
                    self._info.append( row[i] )
        else:
            msg = "Unknown parameter format: '%s'"
            raise ValueError(msg % self.options.format)

    def get_table(self):
        tmp = []
        if not self.options.columns is None:
            tmp.append(self.options.columns)
        if not self.options.set is None:
            # Create column names
            if self.options.columns is None:
                cols = []
                for i in xrange(self.options.set.dimen):
                    cols.append(self.options.set.name+str(i))
                tmp.append(cols)
            # Get rows
            for data in self.options.set:
                if self.options.set.dimen > 1:
                    tmp.append( list(data) )
                else:
                    tmp.append( [data] )
        elif not self.options.param is None:
            if type(self.options.param) in (list,tuple):
                _param = self.options.param
            else:
                _param = [self.options.param]
            tmp = []
            # Collect data
            for index in _param[0]:
                if index is None:
                    row = []
                elif type(index) in (list,tuple):
                    row = list(index)
                else:
                    row = [index]
                for param in _param:
                    row.append( value(param[index]) )
                tmp.append(row)
            # Create column names
            if self.options.columns is None:
                cols = []
                for i in xrange(len(tmp[0])-len(_param)):
                    cols.append('I'+str(i)) 
                for param in _param:
                    cols.append( param)
                tmp = [cols] + tmp
        return tmp




# a brain-dead parabolic function of a single variable, whose minimum is
# obviously at x=0. the entry-level quadratic test-case. the lack of
# constraints could (but shouldn't in a perfect world) cause issues for
# certain solvers.

from coopr.pyomo import *

model = AbstractModel()

model.x = Var(bounds=(-10,10), within=Reals)

def objective_rule(model):
    return model.x * model.x
model.objective = Objective(rule=objective_rule, sense=minimize)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['DataPortal']

import logging
from coopr.pyomo.base.plugin import *

logger = logging.getLogger('coopr.pyomo')

class DataPortal(object):
    """
    An object that manages data imports and exports from
    external data sources.

    This object contains the interface routines for importing and
    exporting data from external data sources.
    """

    def __init__(self, **kwds):
        """
        Constructor
        """
        # Initialize this object with no data manager
        self.data_manager = None

        # maps initialization data as follows: _data[namespace][symbol] -> data
        self._data={}

        # This is the data that is imported from various sources
        self._default={}

        # the model for which this data is associated. used for error
        # checking, e.g., object names must exist in the model, set 
        # dimensions must match, etc.
        self._model = kwds.pop('model', None)

        if 'filename' in kwds:
            filename = kwds.pop('filename')
            self.connect(filename=filename, **kwds)
            self.load()
            self.disconnect()
        elif 'data_dict' in kwds:
            data = kwds.pop('data_dict')
            if not data is None:
                self._data = data
        elif len(kwds) > 0:
            raise ValueError("Unknown options: %s" % str(kwds.keys()))

    #
    # intent of this method is to add, on a component-by-component 
    # basis, initialization data to a ModelData instance. not the
    # "usual" form of data initialization.
    #
    def Xinitialize(self, component, data, namespace=None):
        if namespace not in self._data.keys():
            self._data[namespace]={}
        self._data[namespace][component] = data

    def connect(self, **kwds):
        """
        Construct a data source objects that is associated with the input source.
        This data manager is used to process future data imports and exports.
        """
        if not self.data_manager is None:
            self.data_manager.close()
        data = kwds.get('filename',None)
        if data is None:
            data = kwds.get('server',None)
        tmp = data.split(".")[-1]
        self.data_manager = DataManagerFactory(tmp) 
        if self.data_manager is None:
            from sys import modules
            if 'coopr.environ' not in modules:
                logger.warning(
"""DEPRECATION WARNING: beginning in Coopr 4.0, plugins (including
solvers and DataPortal clients) will not be automatically registered. To
automatically register all plugins bundled with core Coopr, user scripts
should include the line, "import coopr.environ".""" )
                import coopr.environ
                self.connect(**kwds)
                return
            raise IOError("Unknown file format '%s'" % tmp)
        self.data_manager.initialize(**kwds)
        self.data_manager.open()

    def disconnect(self):
        """
        Close the Construct a data source objects that is associated with the input source.
        This data manager is used to process future data imports and exports.
        """
        self.data_manager.close()
        self.data_manager = None
        
    def load(self, **kwds):
        """
        Import data from an external data source.
        """
        _model = kwds.pop('model', None)
        if not _model is None:
            self._model=_model
        if __debug__:
            logger.debug("Loading data...")
        _disconnect=False
        if self.data_manager is None:
            self.connect(**kwds)
            _disconnect=True
        elif len(kwds) > 0:
            self.data_manager.add_options(**kwds)
        self._preprocess()
        if not self.data_manager.read():
            print("Warning: error occured while processing %s" % str(self.data_manager))
        if __debug__:
            logger.debug("Processing data ...")
        status = self.data_manager.process(self._model, self._data, self._default)
        self.data_manager.clear()
        if _disconnect:
            self.disconnect()
        if __debug__:
            logger.debug("Done.")

    def store(self, **kwds):
        """
        Export data from to an external data source.
        """
        if __debug__:
            logger.debug("Storing data...")
        _disconnect=False
        if self.data_manager is None:
            self.connect(**kwds)
            _disconnect=True
        elif len(kwds) > 0:
            self.data_manager.add_options(**kwds)
        self._preprocess()
        self.load_data()
        if not self.data_manager.write(self._data):
            print("Warning: error occured while processing %s" % str(self.data_manager))
        if _disconnect:
            self.disconnect()
        if __debug__:
            logger.debug("Done.")

    def data(self, name, namespace=None):
        if not namespace in self._data:
            raise IOError("Unknown namespace '%s'" % str(namespace))
        ans = self._data[namespace][name]
        if None in ans:
            return ans[None]
        return ans

    def _preprocess(self):
        options = self.data_manager.options
        #
        if options.data is None and (not options.set is None or not options.param is None or not options.index is None):
            options.data = []
            if not options.set is None:
                if type(options.set) in (list,tuple):
                    for item in options.set:
                        options.data.append(item)
                else:
                    options.data.append(options.set)
            if not options.index is None:
                options.data.append(options.index)
            if not options.param is None:
                if type(options.param) in (list,tuple):
                    for item in options.param:
                        options.data.append(item)
                else:
                    options.data.append(options.param)
        #
        if not options.data is None:
            if type(options.data) in (list, tuple):
                ans = []
                for item in options.data:
                    try:
                        ans.append(item.name)
                        self._model = item.model()
                    except:
                        ans.append(item)
                options.data = ans
            else:
                try:
                    self._model = options.data.model()
                    options.data = self.data_manager.options.data.name
                except:
                    pass


    def load_data(self):
        if not self.data_manager.options.data is None:
            if not self._model is None:
                if type(self.data_manager.options.data) is list:
                    for name in self.data_manager.options.data:
                        c = getattr(self._model, name)
                        try:
                            self._data[name] = c.data()
                        except:
                            self._data[name] = c.extract_values()
                else:
                    name = self.data_manager.options.data
                    c = getattr(self._model, name)
                    try:
                        self._data[name] = c.data()
                    except:
                        self._data[name] = c.extract_values()
 



import ast

from coopr.core.plugin import *
from coopr.pyomo.check.hooks import *


class FunctionTrackerHook(SingletonPlugin):

    implements(IPreCheckHook)

    def precheck(self, runner, script, info):
        # create models dict if nonexistent
        if getattr(script, 'functionDefs', None) is None:
            script.functionDefs = {}

        # add new function definitions
        if isinstance(info, ast.FunctionDef):
            script.functionDefs[info.name] = info

        # update function def dictionary with assignments
        elif isinstance(info, ast.Assign):
            if isinstance(info.value, ast.Name):
                if info.value.id in script.functionDefs:
                    for target in info.targets:
                        if isinstance(target, ast.Name):
                            script.functionDefs[target.id] = script.functionDefs[info.value.id]
            else:
                for target in info.targets:
                    if isinstance(target, ast.Name):
                        if target.id in script.functionDefs:
                            del script.functionDefs[target.id]



#
# Imports
#
from coopr.pyomo import *

#
# Setup
#

model = AbstractModel()

model.PROD = Set()

model.STAGE = Set()

model.rate = Param(model.PROD, model.STAGE, within=PositiveReals)

model.avail = Param(model.STAGE, within=NonNegativeReals)

model.profit = Param(model.PROD)

model.commit = Param(model.PROD, within=NonNegativeReals)

model.market = Param(model.PROD, within=NonNegativeReals)

def Make_bounds(model,i):
    return (model.commit[i],model.market[i])
model.Make = Var(model.PROD, bounds=Make_bounds)

def Objective_rule (model):
    return summation(model.profit, model.Make)
model.Total_Profit = Objective(rule=Objective_rule, sense=maximize)

def Timelim_rule(model, s):
    timeexpr = 0
    for p in model.PROD:
        timeexpr = timeexpr + (1.0/model.rate[p,s]) * model.Make[p]
    return timeexpr < model.avail[s]
model.Time = Constraint(model.STAGE, rule=Timelim_rule)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import types
import weakref
from coopr.pyomo.base.numvalue import as_numeric
from coopr.pyomo.base.component import Component
from coopr.pyomo.base.expr import _ExpressionBase, clone_expression, _ExternalFunctionExpression

__all__  = ( 'ExternalFunction', )

try:
    basestring
except:
    basestring = str

class ExternalFunction(Component):
    def __new__(cls, *args, **kwds):
        if cls != ExternalFunction:
            return super(ExternalFunction, cls).__new__(cls)
        if len(args) == 1 and type(args[0]) is types.FunctionType:
            return PythonCallbackFunction.__new__(PythonCallbackFunction)
        elif not args and 'library' not in kwds and 'function' in kwds and \
                type(kwds['function']) is types.FunctionType:
            return PythonCallbackFunction.__new__(PythonCallbackFunction)
        else:
            return BasicExternalFunction.__new__(BasicExternalFunction)

    def __init__(self, *args, **kwds):
        kwds.setdefault('ctype', ExternalFunction)
        Component.__init__(self, **kwds)
        self._constructed = True
        ### HACK ###
        # FIXME: We must declare an _index attribute because
        # block._add_temporary_set assumes ALL components define an
        # index.  Sigh.
        self._index = None

    def __call__(self, *args):
        idxs = range(len(args))
        idxs.reverse()
        for i in idxs:
            if type(args[i]) is types.GeneratorType:
                args = args[:i] + tuple(args[i]) + args[i+1:]
        return _ExternalFunctionExpression( self, tuple(
                _external_fcn__clone_if_needed(x) if isinstance(x, _ExpressionBase)
                else x if isinstance(x, basestring)
                else as_numeric(x)
                for x in args ) )

    def cname(self, fully_qualified=False, name_buffer=None):
        if self.name:
            return super(ExternalFunction, self).cname(
                fully_qualified, name_buffer )
        else:
            return str(self._library) + ":" + str(self._function)
        
    def evaluate(self, args):
        raise NotImplementedError(
            "General external functions can not be evaluated within Python." )


class BasicExternalFunction(ExternalFunction):
    def __init__(self, *args, **kwds):
        if args:
            raise ValueError(
                "BasicExternalFunction constructor does not support "
                "positional arguments" )
        self._library = kwds.pop('library', None)
        self._function = kwds.pop('function', None)
        ExternalFunction.__init__(self, *args, **kwds)



class PythonCallbackFunction(ExternalFunction):
    global_registry = {}

    @classmethod
    def register_instance(instance):
        _id = len(global_registry)
        global_registry[_id] = weakref.ref(instance)
        return _id

    def __init__(self, *args, **kwds):
        if args:
            self._fcn = args[0]
            if len(args) > 1:
                raise ValueError(
                    "PythonCallbackFunction constructor only supports a "
                    "single positional positional arguments" )
        if not args:
            self._fcn = kwds.pop('function')
        if kwds:
            raise ValueError(
                "PythonCallbackFunction constructor does not support "
                "keyword arguments" )
        self._library = 'pyomo_ampl.so'
        self._function = 'pyomo_socket_server'
        ExternalFunction.__init__(self, *args, **kwds)
        self._fcn_id = PythonCallbackFunction.register_instance(self)

    def __call__(self, *args):
        return super(PythonCallbackFunction, self).__call__(self._fcn_id, *args)

    def cname(self, fully_qualified=False, name_buffer=None):
        if self.name:
            return super(ExternalFunction, self).cname(
                fully_qualified, name_buffer )
        else:
            return "PythonCallback(%s)" % str(self._fcn)

    def evaluate(self, args):
        # Skip the library name and function name
        # (we assume that the identifier points to this function!)
        args.next() # library name
        args.next() # function name
        _id = args.next() # global callback function identifyer
        if _id != self._fcn_id:
            raise RuntimeError(
                "PythonCallbackFunction called with invalid Global ID" )
        return self._fcn(*tuple(args))



#
# Unit Tests for Elements of a Model
#
# TestSimpleCon                Class for testing single constraint
# TestArrayCon                Class for testing array of constraint
#

import logging
import os
import sys

from six import StringIO

from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

from coopr.pyomo.base import IntegerSet
from coopr.pyomo import *
from coopr.opt import *
import pyutilib.th as unittest
import pyutilib.services


class TestMutable(unittest.TestCase):
    def test_mutable_constraint_upper(self):
        model = AbstractModel()
        model.P = Param(initialize=2.0, mutable=True)
        model.X = Var()

        def constraint_rule(m):
            return m.X <= m.P
        model.C = Constraint(rule=constraint_rule)

        instance = model.create()

        self.assertEqual(value(instance.C.upper), 2.0)

        instance.P = 4.0
        instance.preprocess()

        self.assertEqual(value(instance.C.upper), 4.0)


    def test_mutable_constraint_lower(self):
        model = AbstractModel()
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var()

        def constraint_rule(m):
            return m.X >= m.Q
        model.C = Constraint(rule=constraint_rule)

        instance = model.create()

        self.assertEqual(value(instance.C.lower), 2.0)

        instance.Q = 4.0
        instance.preprocess()

        self.assertEqual(value(instance.C.lower), 4.0)


    def test_mutable_constraint_both(self):
        model = AbstractModel()
        model.P = Param(initialize=4.0, mutable=True)
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var()

        def constraint_rule(m):
            return m.Q <= m.X <= m.P
        model.C = Constraint(rule=constraint_rule)

        instance = model.create()

        self.assertEqual(value(instance.C.lower), 2.0)
        self.assertEqual(value(instance.C.upper), 4.0)

        instance.P = 8.0
        instance.Q = 1.0
        instance.preprocess()

        self.assertEqual(value(instance.C.lower), 1.0)
        self.assertEqual(value(instance.C.upper), 8.0)



    def test_mutable_var_bounds_lower(self):
        model = AbstractModel()
        model.P = Param(initialize=2.0, mutable=True)
        model.X = Var(bounds=(model.P,None))

        instance = model.create()

        self.assertEqual(instance.X.bounds, (2.0, None))

        instance.P = 4.0
        instance.preprocess()

        self.assertEqual(instance.X.bounds, (4.0, None))


    def test_mutable_var_bounds_upper(self):
        model = AbstractModel()
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var(bounds=(model.Q,None))

        instance = model.create()

        self.assertEqual(instance.X.bounds, (2.0, None))

        instance.Q = 4.0
        instance.preprocess()

        self.assertEqual(instance.X.bounds, (4.0, None))


    def test_mutable_var_bounds_both(self):
        model = AbstractModel()
        model.P = Param(initialize=4.0, mutable=True)
        model.Q = Param(initialize=2.0, mutable=True)
        model.X = Var(bounds=(model.P,model.Q))

        instance = model.create()

        self.assertEqual(value(instance.X.lb), 4.0)
        self.assertEqual(value(instance.X.ub), 2.0)

        instance.P = 8.0
        instance.Q = 1.0
        instance.preprocess()

        self.assertEqual(value(instance.X.lb), 8.0)
        self.assertEqual(value(instance.X.ub), 1.0)

if __name__ == "__main__":
    unittest.main()




# A Suffix example for the gurobi_ampl solver.
#
# This Pyomo example is formulated as a python script.
# To run this script execute the following command:
#
# $ coopr_python gurobi_ampl_example.py
#
# Execution of this script requires that the gurobi_ampl
# solver is in the current search path for executables
# on this system. This example was tested using Gurobi
# Solver 5.0.0

import coopr.environ
from coopr.pyomo import *
from coopr.opt import SolverFactory

### Create the gurobi_ampl solver plugin using the ASL interface
solver = 'gurobi_ampl'
solver_io = 'nl'
stream_solver = False     # True prints solver output to screen
keepfiles =     False     # True prints intermediate file names (.nl,.sol,...) 
opt = SolverFactory(solver,solver_io=solver_io)

if opt is None:
    print("")
    print("ERROR: Unable to create solver plugin for %s "\
          "using the %s interface" % (solver, solver_io))
    print("")
    exit(1)

opt.options['outlev'] = 1 # tell gurobi to be verbose with output
###

### Create a trivial example model
model = ConcreteModel()
model.s = Set(initialize=[1,2,3])
model.x = Var(model.s,within=NonNegativeReals)
model.obj = Objective(expr=summation(model.x))
model.con = Constraint(model.s, rule=lambda model,i: model.x[i] >= i-1)
###

### Declare all suffixes
# The variable solution status suffix
# (this suffix can be sent to the solver and loaded from the solution)
sstatus_table={'bas':1,   # basic
               'sup':2,   # superbasic
               'low':3,   # nonbasic <= (normally =) lower bound
               'upp':4,   # nonbasic >= (normally =) upper bound
               'equ':5,   # nonbasic at equal lower and upper bounds
               'btw':6}   # nonbasic between bounds
model.sstatus = Suffix(direction=Suffix.IMPORT_EXPORT,
                       datatype=Suffix.INT)                       
model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)

# Report the best known bound on the objective function
model.bestbound = Suffix(direction=Suffix.IMPORT)

# A few Gurobi variable solution sensitivity suffixes
model.senslblo = Suffix(direction=Suffix.IMPORT) # smallest variable lower bound
model.senslbhi = Suffix(direction=Suffix.IMPORT) # greatest variable lower bound
model.sensublo = Suffix(direction=Suffix.IMPORT) # smallest variable upper bound
model.sensubhi = Suffix(direction=Suffix.IMPORT) # greatest variable upper bound

# A Gurobi constraint solution sensitivity suffix
model.sensrhshi = Suffix(direction=Suffix.IMPORT) # greatest right-hand side value
###

# Tell gurobi_ampl to report solution sensitivities
# and bestbound via suffixes in the solution file
opt.options['solnsens'] = 1
opt.options['bestbound'] = 1

# Set one of the sstatus suffix values, which will be sent to the solver
model.sstatus.setValue(model.x[1], sstatus_table['low'])

def print_model_suffixes(model):
    # Print all suffix values for all model components in a nice table
    print "\t",
    for name,suffix in active_import_suffix_generator(model):
            print "%8s" % (name,),
    print("")
    for i in model.s:
        print model.x[i],"\t",
        for name,suffix in active_import_suffix_generator(model):
            print "%8s" % (suffix.getValue(model.x[i]),),
        print("")
    for i in model.s:
        print model.con[i],"\t",
        for name,suffix in active_import_suffix_generator(model):
            print "%8s" % (suffix.getValue(model.con[i]),),
        print("") 
    print model.obj.name,"\t",
    for name,suffix in active_import_suffix_generator(model):
        print "%8s" % (suffix.getValue(model.obj),),
    print("") 
    print("") 

print("") 
print("Suffixes Before Solve:")
print_model_suffixes(model)

### Generate the constraint expression trees if necessary
if solver_io != 'nl':
    # only required when not using the ASL interface
    model.preprocess()
###

### Send the model to gurobi_ampl and collect the solution
# The solver plugin will scan the model for all active suffixes
# valid for importing, which it will store into the results object
results = opt.solve(model,
                    keepfiles=keepfiles,
                    tee=stream_solver)
# load the results (including any values for previously declared
# IMPORT / IMPORT_EXPORT Suffix components found in the results object)
model.load(results)
###

print("")
print("Suffixes After Solve:")
print_model_suffixes(model)




# A Suffix example for the gurobi_ampl solver.
#
# This Pyomo example is formulated as a python script.
# To run this script execute the following command:
#
# $ coopr_python gurobi_ampl_iis.py
#
# Execution of this script requires that the gurobi_ampl
# solver is in the current search path for executables
# on this system. This example was tested using Gurobi
# Solver 5.0.0

import coopr.environ
from coopr.pyomo import *
from coopr.opt import SolverFactory

### Create the gurobi_ampl solver plugin using the ASL interface
solver = 'gurobi_ampl'
solver_io = 'nl'
stream_solver = False     # True prints solver output to screen
keepfiles =     False     # True prints intermediate file names (.nl,.sol,...) 
opt = SolverFactory(solver,solver_io=solver_io)

if opt is None:
    print("")
    print("ERROR: Unable to create solver plugin for %s "\
          "using the %s interface" % (solver, solver_io))
    print("")
    exit(1)

# tell gurobi to be verbose with output
opt.options['outlev'] = 1 

# tell gurobi to find an iis table for the infeasible model
opt.options['iisfind'] = 1 # tell gurobi to be verbose with output

### Create a trivial and infeasible example model
model = ConcreteModel()
model.x = Var(within=NonNegativeReals)
model.obj = Objective(expr=model.x)
model.con = Constraint(expr=model.x <= -1)
###

# Create an IMPORT Suffix to store the iis information that will
# be returned by gurobi_ampl
model.iis = Suffix(direction=Suffix.IMPORT)

### Generate the constraint expression trees if necessary
if solver_io != 'nl':
    # only required when not using the ASL interface
    model.preprocess()
###

### Send the model to gurobi_ampl and collect the solution
# The solver plugin will scan the model for all active suffixes
# valid for importing, which it will store into the results object
results = opt.solve(model,
                    keepfiles=keepfiles,
                    tee=stream_solver)
model.load(results)

print("")
print("IIS Results")
for component, value in model.iis.extractValues():
    print component, value



#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly reports the values corresponding
#          to the nl file header line with the label
#          '# nonlinear vars in constraints, objectives, both'
#

from coopr.pyomo import *

model = ConcreteModel()

model.x = Var(initialize=1.0)
model.y = Var(initialize=1.0)

model.OBJ = Objective(expr=model.x*model.y)

model.CON1 = Constraint(expr=model.y**2 == 4)




from coopr.pyomo import *

model = AbstractModel()
model.X = Var()

def c_rule(m):
    if m.X >= 10.0:
        pass
    if value(m.X) >= 10.0:
        pass
    return m.X >= 10.0

model.C = Constraint(rule=c_rule)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

#
# TODO:  Do we need to have different Expression objects for different
#        intrinsic expressions?
# TODO:  If an expression has a fixed value, should we simply call the
#        intrinsic function with the value of that expression?
# TODO:  Do we need to register these expression objects?
#

__all__ = ( 'log', 'log10', 'sin', 'cos', 'tan', 'cosh', 'sinh', 'tanh',
            'asin', 'acos', 'atan', 'exp', 'sqrt', 'asinh', 'acosh', 
            'atanh', 'ceil', 'floor' )

import math

from coopr.pyomo.base.component import Component
from coopr.pyomo.base.expr import _IntrinsicFunctionExpression, \
    _getrefcount_available
from coopr.pyomo.base.numvalue import NumericValue, as_numeric

if _getrefcount_available:
    from sys import getrefcount


def _generate_intrinsic_function_expression__clone_if_needed(obj):
    if getrefcount(obj) - UNREFERENCED_INTRINSIC_EXPR_COUNT == 0:
        return obj
    elif getrefcount(obj) - UNREFERENCED_INTRINSIC_EXPR_COUNT > 0:
        generate_intrinsic_function_expression.clone_counter += 1
        return obj.clone()
    else:
        raise RuntimeError("Expression entered " \
              "generate_intrinsic_function_expression() " \
              "with too few references (%s<0); this is indicative of a " \
              "SERIOUS ERROR in the expression reuse detection scheme." \
              % ( count, ))

def _generate_intrinsic_function_expression__noCloneCheck(obj):
    return obj

def generate_intrinsic_function_expression(args, nargs, name, fcn):
    # Special handling: if there are no Pyomo Modeling Objects in the
    # argument list, then evaluate the expression and return the result.
    pyomo_expression = False
    for arg in args:
        # FIXME: does anyone know why we also test for 'Component' here? [JDS]
        if isinstance(arg, NumericValue) or isinstance(arg, Component):
            # TODO: efficiency: we already know this is a NumericValue -
            # so we should be able to avoid the call to as_numeric()
            # below (expecially since most intrinsic functions are unary
            # operators.
            pyomo_expression = True
            break
    if not pyomo_expression:
        return fcn(*args)

    new_args = []
    for arg in args:
        new_arg = as_numeric(arg)
        if new_arg.is_expression():
            new_arg = _generate_intrinsic_function_expression__clone_if_needed(new_arg)
        elif new_arg.is_indexed():
            raise ValueError("Argument for intrinsic function '%s' is an "\
                "n-ary numeric value: %s\n    Have you given variable or "\
                "parameter '%s' an index?" % (name, new_arg.name, new_arg.name))
        new_args.append(new_arg)
    return _IntrinsicFunctionExpression(name, nargs, tuple(new_args), fcn)

# [debugging] clone_counter is a count of the number of calls to
# expr.clone() made during expression generation.
generate_intrinsic_function_expression.clone_counter = 0

# [configuration] UNREFERENCED_EXPR_COUNT is a "magic number" that
# indicates the stack depth between "normal" modeling and
# _clone_if_needed().  If an expression enters _clone_if_needed() with
# UNREFERENCED_EXPR_COUNT references, then there are no other variables
# that hold a reference to the expression and cloning is not necessary.
# If there are more references than UNREFERENCED_EXPR_COUNT, then we
# must clone the expression before operating on it.  It should be an
# error to hit _clone_if_needed() with fewer than
# UNREFERENCED_EXPR_COUNT references.
UNREFERENCED_INTRINSIC_EXPR_COUNT = 7

#
# If you want to completely disable clone checking (e.g., for
# line-profiling this file), set the following to "if True".
#
if not _getrefcount_available:
    _generate_intrinsic_function_expression__clone_if_needed = \
        _generate_intrinsic_function_expression__noCloneCheck


def fabs(*args):
    return generate_intrinsic_function_expression(args, 1, 'log', math.fabs)

def ceil(*args):
    return generate_intrinsic_function_expression(args, 1, 'ceil', math.ceil)

def floor(*args):
    return generate_intrinsic_function_expression(args, 1, 'floor', math.floor)


# e ** x
def exp(*args):
    return generate_intrinsic_function_expression(args, 1, 'exp', math.exp)

def log(*args):
    return generate_intrinsic_function_expression(args, 1, 'log', math.log)

def log10(*args):
    return generate_intrinsic_function_expression(args, 1, 'log10', math.log10)

# FIXME: this is nominally the same as x ** y, but follows a different path!
def pow(*args):
    return generate_intrinsic_function_expression(args, 1, 'pow', math.pow)

# FIXME: this is nominally the same as x ** 0.5, but follows a different path!
def sqrt(*args):
    return generate_intrinsic_function_expression(args, 1, 'sqrt', math.sqrt)


def sin(*args):
    return generate_intrinsic_function_expression(args, 1, 'sin', math.sin)

def cos(*args):
    return generate_intrinsic_function_expression(args, 1, 'cos', math.cos)

def tan(*args):
    return generate_intrinsic_function_expression(args, 1, 'tan', math.tan)

def sinh(*args):
    return generate_intrinsic_function_expression(args, 1, 'sinh', math.sinh)

def cosh(*args):
    return generate_intrinsic_function_expression(args, 1, 'cosh', math.cosh)

def tanh(*args):
    return generate_intrinsic_function_expression(args, 1, 'tanh', math.tanh)


def asin(*args):
    return generate_intrinsic_function_expression(args, 1, 'asin', math.asin)

def acos(*args):
    return generate_intrinsic_function_expression(args, 1, 'acos', math.acos)

def atan(*args):
    return generate_intrinsic_function_expression(args, 1, 'atan', math.atan)

def asinh(*args):
    return generate_intrinsic_function_expression(args, 1, 'asinh', math.asinh)

def acosh(*args):
    return generate_intrinsic_function_expression(args, 1, 'acosh', math.acosh)

def atanh(*args):
    return generate_intrinsic_function_expression(args, 1, 'atanh', math.atanh)




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Var', 'VarList']

import logging
import sys
import weakref
from six import iteritems, iterkeys, itervalues
from six.moves import xrange

from coopr.pyomo.base.numvalue import NumericValue, value, is_fixed
from coopr.pyomo.base.set_types import BooleanSet, IntegerSet, RealSet, Reals
from coopr.pyomo.base.component import Component, ComponentData, register_component
from coopr.pyomo.base.sparse_indexed_component import SparseIndexedComponent, UnindexedComponent_set, normalize_index
from coopr.pyomo.base.misc import apply_indexed_rule
from coopr.pyomo.base.sets import Set
from coopr.pyomo.base.util import is_functor

logger = logging.getLogger('coopr.pyomo')


_noarg = object()

class _VarData(ComponentData, NumericValue):
    """
    This class defines the data for a single variable.

    Constructor Arguments:
        component   The Var object that owns this data.

    Public Class Attributes:
        domain      The domain of this variable.
        fixed       If True, then this variable is treated as a 
                        fixed constant in the model.
        initial     The default initial value for this variable.
        lb          A lower bound for this variable.  The lower bound can be 
                        either numeric constants, parameter values, expressions 
                        or any object that can be called with no arguments.
        ub          A upper bound for this variable.  The upper bound can be either 
                        numeric constants, parameter values, expressions or any
                        object that can be called with no arguments.
        stale       A Boolean indicating whether the value of this variable is
                        legitimiate.  This value is true if the value should 
                        be considered legitimate for purposes of reporting or 
                        other interrogation. 
        value       The numeric value of this variable.

    The domain, lb, and ub attributes are properties because they
    are too widely accessed directly to enforce explicit getter/setter
    methods and we need to deter directly modifying or accessing
    these attributes in certain cases.
    """

    __pickle_slots__ = ('value','initial','_lb','_ub','fixed','stale')
    __slots__ = __pickle_slots__ + ( '__weakref__', )

    def __init__(self, component):
        """
        Constructor
        """
        #
        # The following is equivalent to calling
        # the base ComponentData constructor, as follows:
        # ComponentData.__init__(self, component)
        #
        self._component = weakref.ref(component)
        #
        # The following is equivalent to calling the
        # base NumericValue constructor, as follows:
        # NumericValue.__init__(self, name, domain, None, False)
        #
        self.value = None
        #
        self.fixed = False
        self.initial = None
        self.stale = True
        #
        # The type of the lower and upper bound attributes can either 
        # be atomic numeric types in Python, expressions, etc. 
        # Basically, they can be anything that passes an "is_fixed" test.
        #
        self._lb = None
        self._ub = None

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        result = super(_VarData, self).__getstate__()
        for i in _VarData.__pickle_slots__:
            result[i] = getattr(self, i)
        return result

    # Note: because NONE of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method.

    #
    # 'domain' is a property so we can ensure that a 'bounds' attribute exists on the
    # domain object.
    #
    @property
    def domain(self):
        """
        Return the domain attribute.
        """
        return self._component()._domain

    #
    # 'lb' is a property because we need to compare it against the lower bound of 
    # the 'domain' attribute to properly handle domains that can change after a 
    # variable has been constructed.
    # 
    @property
    def lb(self):
        """
        Return the lower bound for this variable.
        """
        dlb, _ = self.domain.bounds()
        if self._lb is None:
            return dlb
        elif dlb is None:
            return value(self._lb)
        return max(value(self._lb), dlb)

    @lb.setter
    def lb(self, val):
        """
        Set the lb attribute to the given value.
        """
        raise AttributeError("Assignment not allowed. Use the setlb method")

    def setlb(self, val):
        """
        Set the lower bound for this variable after validating that 
        the value is fixed (or None).
        """
        if val is None:
            self._lb = None
        else:
            if is_fixed(val):
                self._lb = val
            else:
                raise ValueError(
                    "Non-fixed input of type '%s' supplied as variable lower "
                    "bound - legal types must be fixed expressions or variables."
                    % (type(val),) )

    def setub(self, val):
        """
        Set the upper bound for this variable after validating that 
        the value is fixed (or None).
        """
        if val is None:
            self._ub = None
        else:
            if is_fixed(val):
                self._ub = val
            else:
                raise ValueError(
                    "Non-fixed input of type '%s' supplied as variable upper "
                    "bound - legal types are fixed expressions or variables."
                    "parameters"
                    % (type(val),) )

    #
    # 'ub' is a property because we need to compare it against the upper bound of
    # the 'domain' attribute to properly handle domains that can change after a
    # variable has been constructed.
    # 
    @property
    def ub(self):
        """
        Return the upper bound for this variable.
        """
        _, dub = self.domain.bounds()
        if self._ub is None:
            return dub
        elif dub is None:
            return value(self._ub)
        return min(value(self._ub), dub)

    @ub.setter
    def ub(self, val):
        """
        Set the ub attribute to the given value.
        """
        raise AttributeError("Assignment not allowed. Use the setub method")

    #
    # 'bounds' is a property because we need to compare it against the bounds of 
    # the 'domain' attribute to properly handle domains that can change after a
    # variable has been constructed.
    # 
    @property
    def bounds(self):
        """
        Returns the tuple (lower bound, upper bound).
        """
        return (self.lb, self.ub)

    @bounds.setter
    def bounds(self, val):
        """
        Set the bounds attribute to the given value.
        """
        raise AttributeError("Assignment not allowed. Use the setub and setlb methods")

    def __call__(self, exception=True):
        """
        Return the value of this object.
        """
        return self.value

    def is_integer(self):
        """
        Returns True when the domain is an instance of IntegerSet.
        """
        return self.domain.__class__ is IntegerSet

    def is_binary(self):
        """
        Returns True when the domain is an instance of BooleanSet.
        """
        return self.domain.__class__ is BooleanSet

    def is_continuous(self):
        """
        Returns True when the domain is an instance of RealSet.
        """
        return self.domain.__class__ is RealSet

    def fix(self, val=_noarg):
        """
        Set the fixed indicator to True. Value argument is optional,
        indicating the variable should be fixed at its current value.

        We use the global _noarg object as the default value to allow
        fixing variables to None.
        """
        self.fixed = True
        if val is not _noarg:
            self.value = val

    def unfix(self):
        """
        Sets the fixed indicator to False.
        """
        self.fixed = False

    free=unfix

    def is_fixed(self):
        """
        Returns True if this variable is fixed, otherwise returns False.
        """
        if self.fixed:
            return True
        return False

    def is_constant(self):
        """
        Returns False because this is not a constant in an expression.
        """
        return False

    def polynomial_degree(self):
        """
        If the variable is fixed, it represents a constant is a polynomial with degree 0.
        Otherwise, it has degree 1.

        This method is used in expressions to compute polynomial degree.
        """
        if self.fixed:
            return 0
        return 1

    def set_value(self, val, valid=False):
        """
        Set the value of this numeric object, after validating its value.

        If the 'valid' flag is True, then the validation step is skipped.
        """
        if valid or self._valid_value(val):
            self.value = val
            self.stale = False

    def _valid_value(self, val, use_exception=True):
        """
        Validate the value.  If use_exception is True, then raise an
        exception.
        """
        ans = val is None or val in self.domain
        if not ans and use_exception:
            raise ValueError("Numeric value `%s` is not in domain %s" % (val, self.domain))
        return ans

    def clear(self):
        self.value = None

    def __nonzero__(self):
        """
        Return True if the value is defined and non-zero.
        """
        if self.value:
            return True
        if self.value is None:
            raise ValueError("Var value is undefined")
        return False

    __bool__ = __nonzero__


class _VarDataWithDomain(_VarData):
    """
    Variable data that contains a local domain.
    """

    __slots__ = ( '_domain', )

    def __init__(self, component):
        _VarData.__init__(self, component)
        self._domain = Reals

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        result = super(_VarDataWithDomain, self).__getstate__()
        for i in _VarDataWithDomain.__slots__:
            result[i] = getattr(self, i)
        return result

    #
    # 'domain' is a property so we can ensure that a 'bounds' attribute exists on the
    # domain object.
    #
    @property
    def domain(self):
        """
        Return the domain attribute.
        """
        return self._domain

    @domain.setter
    def domain(self, domain):
        """
        Set the domain attribute to the given value.
        """
        if hasattr(domain, 'bounds'):
            self._domain = domain
        else:
            raise ValueError("%s is not a valid domain. Variable domains must be an instance of "
                             "one of %s an object that declares a method for bounds (like a Pyomo Set)."
                             "Examples: NonNegativeReals, Integers, Binary"
                             % (domain, (RealSet, IntegerSet, BooleanSet)))


class Var(SparseIndexedComponent):
    """
    A numeric variable, which may be defined over an index.

    Constructor Arguments:
        name        The name of this variable
        index       The index set that defines the distinct variables.
                        By default, this is None, indicating that there
                        is a single variable.
        domain      A set that defines the type of values that
                        each variable must be.
        bounds      A rule for defining bounds values for this
                        variable.
        rule        A rule for setting up this variable with
                        existing model data
    """

    def __new__(cls, *args, **kwds):
        if cls != Var:
            return super(Var, cls).__new__(cls)
        if args == ():
            return SimpleVar.__new__(SimpleVar)
        else:
            domain = kwds.get('within', Reals )
            domain = kwds.get('domain', domain )
            if is_functor(domain):
                return IndexedVar.__new__(IndexedVar)
            else:
                return IndexedVarWithDomain.__new__(IndexedVarWithDomain)

    def __init__(self, *args, **kwd):
        #
        # Default keyword values
        #
        initialize = kwd.pop('initialize', None )
        domain = kwd.pop('within', Reals )
        domain = kwd.pop('domain', domain )
        bounds = kwd.pop('bounds', None )
        #
        # Initialize the base class
        #
        kwd.setdefault('ctype', Var)
        SparseIndexedComponent.__init__(self, *args, **kwd)
        #
        # Determine if the domain argument is a functor or other object
        #
        self._domain_init_value = None
        self._domain_init_rule = None
        if is_functor(domain):
            self._domain_init_rule = domain
            self._VarData = _VarDataWithDomain
        else:
            self._domain_init_value = domain
            self._VarData = _VarData
        #
        # Allow for functions or functors for value initialization,
        # without confusing with Params, etc (which have a __call__ method).
        #
        self._value_init_value = None
        self._value_init_rule = None
        if  is_functor(initialize) and (not isinstance(initialize,NumericValue)):
            self._value_init_rule = initialize
        else:
            self._value_init_value = initialize
        #
        # Determine if the bound argument is a functor or other object
        #
        self._bounds_init_rule = None
        self._bounds_init_value = None
        if is_functor(bounds):
            self._bounds_init_rule = bounds
        elif type(bounds) is tuple:
            self._bounds_init_value = bounds
        elif bounds is not None:
            raise ValueError("Variable 'bounds' keyword must be a tuple or function")

    def as_numeric(self):
        """
        Return the current object, which we treat as a numeric object.
        """
        return self

    def is_expression(self):
        """
        Return True if this numeric value is an expression.
        """
        return False

    def is_relational(self):
        """
        Return True if this numeric value represents a relational expression.
        """
        return False

    def flag_as_stale(self):
        """
        Set the 'stale' attribute of every variable data object to True.
        """
        for var_data in itervalues(self._data):
            var_data.stale = True

    def get_values(self, include_fixed_values=True):
        """
        Return a dictionary of index-value pairs.
        """
        if include_fixed_values:
            return dict((idx, vardata.value) 
                            for idx, vardata in iteritems(self._data))
        return dict((idx, vardata.value) 
                            for idx, vardata in iteritems(self._data) 
                                                if not vardata.fixed)

    def set_values(self, new_values, valid=False):
        """
        Set the values of a dictionary.

        The default behavior is to validate the values in the
        dictionary.
        """
        for index, new_value in iteritems(new_values):
            self._data[index].set_value(new_value, valid)

    def reset(self):
        """
        Reset the variable values to their initial values.
        """
        for vardata in itervalues(self):
            vardata.set_value(vardata.initial)

    def __setitem__(self, ndx, val):
        """
        Define the setitem operation:
            var[ndx] = val
        """
        #
        # Get the variable data object
        #
        if ndx in self._data:
            vardata = self._data[ndx]
        else:
            _ndx = normalize_index(ndx)
            if _ndx in self._data:
                vardata = self._data[_ndx]
            else:
                msg = "Cannot set the value of variable '%s' with invalid " \
                    "index '%s'"
                raise KeyError(msg % ( self.cname(True), str(ndx) ))
        #  
        # Set the value
        #
        vardata.set_value(val)

    def construct(self, data=None):
        """
        Initialize this component
        """
        if __debug__:   #pragma:nocover
            try:
                name = str(self.cname(True))
            except:
                # Some Var components don't have a name yet, so just use the type
                name = type(self)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing Variable, name=%s, from data=%s", name, str(data))
        #
        if self._constructed:
            return
        self._constructed=True
        #
        # Construct self._VarData objects for all index values
        #
        if self.is_indexed():
            self._add_members(self._index)
        self._initialize_members(self._index)

    def add(self, index):
        """
        Add a variable with a particular index
        """
        self._data[index] = self._VarData(self)
        return self._data[index]

    def _add_members(self, init_set):
        """
        Create variable data for all indices in a set
        """
        self._data.update((ndx,self._VarData(self)) for ndx in init_set)

    def _initialize_members(self, init_set):
        """
        Initialize variable data for all indices in a set
        """
        #
        # Initialize domains
        #
        if self._domain_init_rule is not None:
            #
            # Initialize domains with a rule
            #
            if self.is_indexed():
                for ndx in init_set:
                    self._data[ndx].domain = apply_indexed_rule(self, self._domain_init_rule, self._parent(), ndx)
            else:
                self.domain = self._domain_init_rule(self._parent())
        else:
            #
            # Initialize domains with a value
            #
            self._domain = self._domain_init_value
        #
        # Initialize values
        #
        if self._value_init_value is not None:
            #
            # Initialize values with a value
            #
            if self._value_init_value.__class__ is dict:
                for key in init_set:
                    #
                    # Skip indices that are not in the dictionary.  This arises when
                    # initializing VarList objects with a dictionary.
                    #
                    if not key in self._value_init_value:
                        continue
                    val = self._value_init_value[key]
                    vardata = self._data[key]
                    vardata.set_value(val)
                    vardata.initial = val
            else:
                val = value(self._value_init_value)
                for key in init_set:
                    vardata = self._data[key]
                    vardata.set_value(val)
                    vardata.initial = val
        elif self._value_init_rule is not None:
            #
            # Initialize values with a rule
            #
            if self.is_indexed():
                for key in init_set:
                    vardata = self._data[key]
                    val = apply_indexed_rule( self, self._value_init_rule,
                                              self._parent(), key )
                    val = value(val)
                    vardata.set_value(val)
                    vardata.initial = val
            else:
                val = self._value_init_rule(self._parent())
                val = value(val)
                self.set_value(val)
                self.initial = val
        #
        # Initialize bounds
        #
        if self._bounds_init_rule is not None:
            #
            # Initialize bounds with a rule
            #
            if self.is_indexed():
                for key in init_set:
                    vardata = self._data[key]
                    (lb, ub) = apply_indexed_rule( self, self._bounds_init_rule,
                                                   self._parent(), key )
                    vardata.setlb(lb)
                    vardata.setub(ub)
            else:
                (lb, ub) = self._bounds_init_rule(self._parent())
                self.setlb(lb)
                self.setub(ub)
        elif self._bounds_init_value is not None:
            #
            # Initialize bounds with a value
            #
            (lb, ub) = self._bounds_init_value
            for key in init_set:
                vardata = self._data[key]
                vardata.setlb(lb)
                vardata.setub(ub)

    def _pprint(self):
        """
        Print component information.
        """
        return ( [("Size", len(self)),
                  ("Index", self._index \
                       if self._index != UnindexedComponent_set else None),
                  ("Domain", None if self._domain_init_value is None else self._domain_init_value.name),
                  ],
                 iteritems(self._data),
                 ( "Key","Lower","Value","Upper","Initial","Fixed","Stale" ),
                 lambda k, v: [ k,
                                value(v.lb),
                                v.value,
                                value(v.ub),
                                v.initial,
                                v.fixed,
                                v.stale,
                                ]
                 )


class SimpleVar(_VarDataWithDomain, Var):
    """
    _VarObject is the implementation representing a single, non-indexed
    variable.
    """

    def __init__(self, *args, **kwd):
        _VarDataWithDomain.__init__(self, self)
        Var.__init__(self, *args, **kwd)
        self._data[None] = self

    #
    # Since this class derives from Component and Component.__getstate__
    # just packs up the entire __dict__ into the state dict, we do not
    # need to define the __getstate__ or __setstate__ methods.
    # We just defer to the super() get/set state.  Since all of our 
    # get/set state methods rely on super() to traverse the MRO, this 
    # will automatically pick up both the Component and Data base classes.
    #

    def __call__(self, exception=True):
        """
        Return the value of this variable.
        """
        if self._constructed:
            return _VarDataWithDomain.__call__(self, exception=exception)
        if exception:
            raise ValueError( """Evaluating the numeric value of variable '%s' before the Var has been
            constructed (there is currently no value to return).""" % self.cname(True) )

    def reset(self):
        """
        Reset the variable values to their initial values.
        """
        self.set_value(self.initial)


class IndexedVar(Var):
    """
    An array of variables.
    """
    
    def __call__(self, exception=True):
        """Compute the value of the variable"""
        if exception:
            msg = 'Cannot compute the value of an array of variables'
            raise TypeError(msg)

    def fix(self, val=_noarg):
        """
        Set the fixed indicator to True. Value argument is optional,
        indicating the variable should be fixed at its current value.

        We use the global _noarg object as the default value to allow
        fixing variables to None.
        """
        for vardata in itervalues(self):
            vardata.fix(val)

    def unfix(self):
        """
        Sets the fixed indicator to False.
        """
        for vardata in itervalues(self):
            vardata.unfix()

    free=unfix

    #
    # 'domain' is a property so we can ensure that a 'bounds' attribute exists on the
    # domain object.
    #
    @property
    def domain(self):
        """
        Return the domain attribute.
        """
        return None


class IndexedVarWithDomain(IndexedVar):
    """
    An array of variables.
    """
    
    def __init__(self, *args, **kwds):
        IndexedVar.__init__(self, *args, **kwds)
        self._domain = None

    #
    # 'domain' is a property so we can ensure that a 'bounds' attribute exists on the
    # domain object.
    #
    @property
    def domain(self):
        """
        Return the domain attribute.
        """
        return self._domain

    @domain.setter
    def domain(self, domain):
        """
        Set the domain attribute to the given value.
        """
        if hasattr(domain, 'bounds'):
            self._domain = domain
        else:
            raise ValueError("%s is not a valid domain. Variable domains must be an instance of "
                             "one of %s an object that declares a method for bounds (like a Pyomo Set)."
                             "Examples: NonNegativeReals, Integers, Binary"
                             % (domain, (RealSet, IntegerSet, BooleanSet)))


class VarList(IndexedVar):
    """
    Variable-length indexed variable objects used to construct Pyomo models.
    """

    End = ( 1003, )

    def __new__(cls, *args, **kwds):
        if cls != VarList:
            return super(VarList, cls).__new__(cls)
        domain = kwds.get('within', Reals )
        domain = kwds.get('domain', domain )
        if is_functor(domain):
            return _VarList.__new__(_VarList)
        else:
            return _VarListWithDomain.__new__(_VarListWithDomain)

    def __init__(self, *args, **kwds):
        IndexedVar.__init__(self, *args, **kwds)


class _VarList(VarList):
    """
    Variable-length indexed variable objects, with local domain attribute.
    """

    def __init__(self, *args, **kwds):
        VarList.__init__(self, *args, **kwds)

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            raise ValueError("Cannot specify indices for a VarList object")
        self._hidden_index = Set()
        IndexedVar.__init__(self, self._hidden_index, **kwargs)
        self._nvars = 0

    def construct(self, data=None):
        """
        Initialize this component
        """
        if __debug__:
            logger.debug("Constructing variable list %s",self.cname(True))
        self._hidden_index.construct()
        IndexedVar.construct(self, data)

    def add(self):
        """
        Add a variable to this list
        """
        index = self._nvars
        self._hidden_index.add(index)
        vardata = self._data[index] = self._VarData(self)
        self._initialize_members([index])
        self._nvars += 1
        return vardata


class _VarListWithDomain(_VarList):
    """
    Variable-length indexed variable objects, with local domain attribute.
    """

    def __init__(self, *args, **kwds):
        _VarList.__init__(self, *args, **kwds)
        self._domain = None

    #
    # 'domain' is a property so we can ensure that a 'bounds' attribute exists on the
    # domain object.
    #
    @property
    def domain(self):
        """
        Return the domain attribute.
        """
        return self._domain

    @domain.setter
    def domain(self, domain):
        """
        Set the domain attribute to the given value.
        """
        if hasattr(domain, 'bounds'):
            self._domain = domain
        else:
            raise ValueError("%s is not a valid domain. Variable domains must be an instance of "
                             "one of %s an object that declares a method for bounds (like a Pyomo Set)."
                             "Examples: NonNegativeReals, Integers, Binary"
                             % (domain, (RealSet, IntegerSet, BooleanSet)))


register_component(Var, "Decision variables in a model.")
register_component(VarList, "Variable-length list of decision variables in a model.")




#
# Test behavior of concrete classes.
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
import pyutilib.services
from coopr.pyomo import *
from coopr.opt import *
import coopr.environ

solver = load_solvers('glpk')


@unittest.skipIf(solver['glpk'] is None, "glpk solver is not available")
class Test(unittest.TestCase):

    def test_blending(self):
        """ The blending example from the PuLP documentation """
        model = AbstractModel()
        model.concrete_mode()

        model.x1 = Var(bounds=(0,None), doc="ChickenPercent")
        model.x2 = Var(bounds=(0,None), doc="BeefPercent")

        model.obj = Objective(expr=0.013*model.x1 + 0.008*model.x2, doc="Total Cost of Ingredients per can")

        model.c0 = Constraint(expr=model.x1+model.x2 == 100.0, doc="Percentage Sum")
        model.c1 = Constraint(expr=0.100*model.x1 + 0.200*model.x2 >= 8.0, doc="Protein Requirement")
        model.c2 = Constraint(expr=0.080*model.x1 + 0.100*model.x2 >= 6.0, doc="Fat Requirement")
        model.c3 = Constraint(expr=0.001*model.x1 + 0.005*model.x2 <= 2.0, doc="Fiber Requirement")
        model.c4 = Constraint(expr=0.002*model.x1 + 0.005*model.x2 <= 0.4, doc="Salt Requirement")

        instance = model.create()
        opt = solver['glpk']
        results = opt.solve(instance)
        new_results = instance.update_results(results)
        new_results.write(filename=currdir+"blend.out", format='json')
        self.assertMatchesJsonBaseline(currdir+"blend.out",currdir+"blend.txt", tolerance=1e-2)


if __name__ == "__main__":
    unittest.main()



#
# Imports
#
from coopr.pyomo import *

#
# Setup
#

model = AbstractModel()

# ***********************************

model.PROD = Set(doc='products')

model.ACT = Set(doc='activities')

# ***********************************

model.cost = Param(model.ACT, within=PositiveReals, doc='cost per unit of each activity')

model.demand = Param(model.PROD, within=NonNegativeReals, doc='units of demand for each product')

model.io = Param(model.PROD, model.ACT, within=NonNegativeReals, doc='units of each product from 1 unit of each activity')

model.level_min = Param(model.ACT, within=NonNegativeReals, doc='min allowed level for each activity')

model.level_max = Param(model.ACT, within=NonNegativeReals, doc='max allowed level for each activity')

# ***********************************

def Level_bounds(model, i):
    return (model.level_min[i], model.level_max[i])
model.Level = Var(model.ACT, bounds=Level_bounds, doc='level for each activity')

# ***********************************

def Total_Cost_rule(model):
    return summation(model.cost, model.Level)
model.Total_Cost = Objective(rule=Total_Cost_rule, doc='minimize total cost')

def Demand_rule(model, i):
    expr = 0
    for j in model.ACT:
        expr += model.io[i,j] * model.Level[j]
    return model.demand[i] < expr
model.Demand = Constraint(model.PROD, rule=Demand_rule, doc='total level for each activity exceeds demand')
